# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import warnings
from enum import Enum, auto
from functools import total_ordering
from typing import Optional

from spdx import utils
from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.license import License
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.parsers.loggers import ErrorMessages


class FileType(Enum):
    SOURCE = auto()
    BINARY = auto()
    ARCHIVE = auto()
    OTHER = auto()
    APPLICATION = auto()
    AUDIO = auto()
    IMAGE = auto()
    TEXT = auto()
    DOCUMENTATION = auto()
    SPDX = auto()
    VIDEO = auto()


def file_type_from_rdf(rdf_file_type: str) -> FileType:
    """e.g. convert fileType_source to FileType.SOURCE"""
    file_type_str = rdf_file_type.split("_")[1].upper()

    if file_type_str not in FileType.__members__:
        raise SPDXValueError("File:FileType")

    return FileType[file_type_str]


def file_type_to_rdf(file_type: FileType) -> str:
    """e.g. convert SOURCE to fileType_source"""
    return f"fileType_{file_type.name.lower()}"


@total_ordering
class File(object):
    """
    Represent an SPDX file.
    Fields:
    - name: File name, str mandatory one.
    - spdx_id: Uniquely identify any element in an SPDX document which may be
    referenced by other elements. Mandatory, one. Type: str.
    - comment: File comment str, Optional zero or one.
    - file_types: list of file types. Cardinality 0..*
    - checksums: Dict with checksum.ChecksumAlgorithm as key and checksum.Checksum as value,
    there must be a SHA1 hash, at least.
    - conc_lics: Mandatory one. license.License or utils.NoAssert or utils.SPDXNone.
    - licenses_in_file: list of licenses found in file, mandatory one or more.
      document.License or utils.SPDXNone or utils.NoAssert.
    - document.license or utils.NoAssert or utils.SPDXNone.
    - license_comment: Optional.
    - copyright: Copyright text, Mandatory one. utils.NoAssert or utils.SPDXNone or str.
    - notice: optional One, str.
    - contributors: List of strings.
    - dependencies: list of file locations.
    - artifact_of_project_name: list of project names, possibly empty.
    - artifact_of_project_home: list of project home page, possibly empty.
    - artifact_of_project_uri: list of project uris, possibly empty.
    -attribution_text: optional string.
    """

    def __init__(self, name, spdx_id=None):
        self.name = name
        self.spdx_id = spdx_id
        self.comment = None
        self.file_types = []
        self.checksums = {}
        self.conc_lics = None
        self.licenses_in_file = []
        self.license_comment = None
        self.copyright = None
        self.notice = None
        self.attribution_text = None
        self.contributors = []
        self.dependencies = []
        self.artifact_of_project_name = []
        self.artifact_of_project_home = []
        self.artifact_of_project_uri = []

    def __eq__(self, other):
        return isinstance(other, File) and self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    @property
    def checksum(self):
        """
        Backwards compatibility, return SHA1 checksum.
        """
        warnings.warn("This property is deprecated. Use get_checksum instead.")
        return self.get_checksum(ChecksumAlgorithm.SHA1)

    @checksum.setter
    def checksum(self, value):
        """
        Backwards compatibility, set checksum.
        """
        warnings.warn("This property is deprecated. Use set_checksum instead.")
        if isinstance(value, str):
            self.set_checksum(Checksum("SHA1", value))
        elif isinstance(value, Checksum):
            self.set_checksum(value)

    def add_lics(self, lics):
        self.licenses_in_file.append(lics)

    def add_contrib(self, contrib):
        self.contributors.append(contrib)

    def add_depend(self, depend):
        self.dependencies.append(depend)

    def add_artifact(self, symbol, value):
        """
        Add value as artifact_of_project{symbol}.
        """
        symbol = "artifact_of_project_{}".format(symbol)
        artifact = getattr(self, symbol)
        artifact.append(value)

    def validate(self, messages):
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        messages.push_context(self.name)
        self.validate_concluded_license(messages)
        self.validate_file_types(messages)
        self.validate_checksums(messages)
        self.validate_licenses_in_file(messages)
        self.validate_copyright(messages)
        self.validate_artifacts(messages)
        self.validate_spdx_id(messages)
        messages.pop_context()
        return messages

    def validate_spdx_id(self, messages):
        if self.spdx_id is None:
            messages.append("File has no SPDX Identifier.")

        return messages

    def validate_copyright(self, messages):
        if self.copyright and not isinstance(
            self.copyright,
            (str, utils.NoAssert, utils.SPDXNone),
        ):
            messages.append(
                "File copyright must be str or unicode or "
                "spdx.utils.NoAssert or spdx.utils.SPDXNone"
            )

        return messages

    def validate_artifacts(self, messages):
        if len(self.artifact_of_project_home) < max(
            len(self.artifact_of_project_uri), len(self.artifact_of_project_name)
        ):
            messages.append(
                "File must have as much artifact of project as uri or homepage"
            )

        return messages

    def validate_licenses_in_file(self, messages):
        for license_in_file in self.licenses_in_file:
            if not isinstance(
                license_in_file, (utils.SPDXNone, utils.NoAssert, License)
            ):
                messages.append(
                    "License in file must be instance of "
                    "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                    "spdx.license.License"
                )

        return messages

    def validate_concluded_license(self, messages):
        if self.conc_lics and not isinstance(
            self.conc_lics, (utils.SPDXNone, utils.NoAssert, License)
        ):
            messages.append(
                "File concluded license must be instance of "
                "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                "spdx.license.License"
            )

        return messages

    def validate_file_types(self, messages):
        for file_type in self.file_types:
            if not isinstance(file_type, FileType):
                messages.append(f"{file_type} is not of type FileType.")
        return messages

    def validate_checksums(self, messages: ErrorMessages):
        for checksum in self.checksums.values():
            if not isinstance(checksum, Checksum):
                messages.append("File checksum must be instance of spdx.checksum.Checksum.")

        if self.get_checksum(ChecksumAlgorithm.SHA1) is None:
            messages.append("At least one file checksum algorithm must be SHA1")

    def calculate_checksum(self, hash_algorithm='SHA1'):
        if hash_algorithm not in ChecksumAlgorithm.__members__:
            raise ValueError
        BUFFER_SIZE = 65536

        file_hash = hashlib.new(hash_algorithm.lower())
        with open(self.name, "rb") as file_handle:
            while True:
                data = file_handle.read(BUFFER_SIZE)
                if not data:
                    break
                file_hash.update(data)

        return file_hash.hexdigest()

    def get_checksum(self, hash_algorithm: ChecksumAlgorithm = ChecksumAlgorithm.SHA1) -> Optional[Checksum]:
        return self.checksums.get(hash_algorithm)

    def set_checksum(self, new_checksum: Checksum):
        if not isinstance(new_checksum, Checksum):
            raise SPDXValueError

        self.checksums[new_checksum.identifier] = new_checksum

    def has_optional_field(self, field):
        return bool(getattr(self, field, None))



# Copyright (c) spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import spdx.file as spdxfile
from spdx.parsers import jsonparser
from spdx.parsers import yamlparser
from spdx.parsers import rdf
from spdx.parsers import xmlparser
from spdx.parsers import tagvalue
from spdx.parsers.loggers import StandardLogger
from spdx.parsers import jsonyamlxmlbuilders, tagvaluebuilders, rdfbuilders
from spdx.parsers.builderexceptions import FileTypeError


def parse_file(fn, encoding="utf-8"):
    builder_module = jsonyamlxmlbuilders
    read_data = False
    if fn.endswith(".rdf") or fn.endswith(".rdf.xml"):
        encoding = None
        parsing_module = rdf
        builder_module = rdfbuilders
    elif fn.endswith(".tag") or fn.endswith(".spdx"):
        parsing_module = tagvalue
        builder_module = tagvaluebuilders
        read_data = True
    elif fn.endswith(".json"):
        parsing_module = jsonparser
    elif fn.endswith(".xml"):
        parsing_module = xmlparser
    elif fn.endswith(".yaml") or fn.endswith(".yml"):
        parsing_module = yamlparser
    else:
        raise FileTypeError("FileType Not Supported" + str(fn))

    p = parsing_module.Parser(builder_module.Builder(), StandardLogger())
    if hasattr(p, "build"):
        p.build()
    with open(fn, "r", encoding=encoding) as f:
        if read_data:
            data = f.read()
            return p.parse(data)
        else:
            return p.parse(f)



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from datetime import datetime
from unittest import TestCase

import spdx
from spdx import utils
from spdx.package import PackagePurpose
from spdx.parsers.tagvalue import Parser
from spdx.parsers.lexers.tagvalue import Lexer
from spdx.parsers.tagvaluebuilders import Builder
from spdx.parsers.loggers import StandardLogger
from spdx.version import Version

document_str = '\n'.join([
        'SPDXVersion: SPDX-2.1',
        'DataLicense: CC0-1.0',
        'DocumentName: Sample_Document-V2.1',
        'SPDXID: SPDXRef-DOCUMENT',
        'DocumentComment: <text>Sample Comment</text>',
        'DocumentNamespace: https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301'
    ])

creation_str = '\n'.join([
    'Creator: Person: Bob (bob@example.com)',
    'Creator: Organization: Acme.',
    'Created: 2010-02-03T00:00:00Z',
    'CreatorComment: <text>Sample Comment</text>'
])

review_str = '\n'.join([
    'Reviewer: Person: Bob the Reviewer',
    'ReviewDate: 2010-02-10T00:00:00Z',
    'ReviewComment: <text>Bob was Here.</text>',
    'Reviewer: Person: Alice the Reviewer',
    'ReviewDate: 2011-02-10T00:00:00Z',
    'ReviewComment: <text>Alice was also here.</text>'
])

package_str = '\n'.join([
    'PackageName: Test',
    'SPDXID: SPDXRef-Package',
    'PackageVersion: 1:22.36.1-8+deb11u1',
    'PackageDownloadLocation: http://example.com/test',
    'FilesAnalyzed: True',
    'PackageSummary: <text>Test package</text>',
    'PackageSourceInfo: <text>Version 1.0 of test</text>',
    'PackageFileName: test-1.0.zip',
    'PackageSupplier: Organization:ACME',
    'PackageOriginator: Organization:ACME',
    'PackageChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12',
    'PackageVerificationCode: 4e3211c67a2d28fced849ee1bb76e7391b93feba (something.rdf, something.txt)',
    'PackageDescription: <text>A package.</text>',
    'PackageComment: <text>Comment on the package.</text>',
    'PackageCopyrightText: <text> Copyright 2014 Acme Inc.</text>',
    'PackageLicenseDeclared: Apache-2.0',
    'PackageLicenseConcluded: (LicenseRef-2.0 and Apache-2.0)',
    'PackageLicenseInfoFromFiles: Apache-1.0',
    'PackageLicenseInfoFromFiles: Apache-2.0',
    'PackageLicenseComments: <text>License Comments</text>',
    'ExternalRef: SECURITY cpe23Type cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:',
    'ExternalRefComment: <text>Some comment about the package.</text>',
    'PrimaryPackagePurpose: OPERATING-SYSTEM',
    'BuiltDate: 2020-01-01T12:00:00Z',
    'ReleaseDate: 2021-01-01T12:00:00Z',
    'ValidUntilDate: 2022-01-01T12:00:00Z'
])

file_str = '\n'.join([
    'FileName: testfile.java',
    'SPDXID: SPDXRef-File',
    'FileType: SOURCE',
    'FileType: TEXT',
    'FileChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12',
    'LicenseConcluded: Apache-2.0',
    'LicenseInfoInFile: Apache-2.0',
    'FileCopyrightText: <text>Copyright 2014 Acme Inc.</text>',
    'ArtifactOfProjectName: AcmeTest',
    'ArtifactOfProjectHomePage: http://www.acme.org/',
    'ArtifactOfProjectURI: http://www.acme.org/',
    'FileComment: <text>Very long file</text>',
    'FileAttributionText: <text>Acknowledgements that might be required to be communicated in some contexts.</text>'
        ])

unknown_tag_str = 'SomeUnknownTag: SomeUnknownValue'

snippet_str = '\n'.join([
    'SnippetSPDXID: SPDXRef-Snippet',
    'SnippetLicenseComments: <text>Some lic comment.</text>',
    'SnippetCopyrightText: <text> Copyright 2008-2010 John Smith </text>',
    'SnippetComment: <text>Some snippet comment.</text>',
    'SnippetName: from linux kernel',
    'SnippetFromFileSPDXID: SPDXRef-DoapSource',
    'SnippetLicenseConcluded: Apache-2.0',
    'LicenseInfoInSnippet: Apache-2.0',
    'SnippetByteRange: 310:420  ',
    'SnippetLineRange: 5:7',
])

annotation_str = '\n'.join([
    'Annotator: Person: Jane Doe()',
    'AnnotationDate: 2010-01-29T18:30:22Z',
    'AnnotationComment: <text>Document level annotation</text>',
    'AnnotationType: OTHER',
    'SPDXREF: SPDXRef-DOCUMENT'
])

document_str_with_empty_line = "\n".join(
    ['SPDXVersion: SPDX-2.1', '  ', 'DataLicense: CC0-1.0'])


class TestLexer(TestCase):
    maxDiff = None

    def setUp(self):
        self.l = Lexer()
        self.l.build()

    def test_document(self):
        data = document_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'DOC_VERSION', 'SPDXVersion', 1)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDX-2.1', 1)
        self.token_assert_helper(self.l.token(), 'DOC_LICENSE', 'DataLicense', 2)
        self.token_assert_helper(self.l.token(), 'LINE', 'CC0-1.0', 2)
        self.token_assert_helper(self.l.token(), 'DOC_NAME', 'DocumentName', 3)
        self.token_assert_helper(self.l.token(), 'LINE', 'Sample_Document-V2.1',
                                 3)
        self.token_assert_helper(self.l.token(), 'SPDX_ID', 'SPDXID', 4)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDXRef-DOCUMENT', 4)
        self.token_assert_helper(self.l.token(), 'DOC_COMMENT', 'DocumentComment', 5)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Sample Comment</text>', 5)
        self.token_assert_helper(self.l.token(), 'DOC_NAMESPACE',
                                 'DocumentNamespace', 6)
        self.token_assert_helper(self.l.token(), 'LINE',
                                 'https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301',
                                 6)

    def test_external_document_references(self):
        data = '''
        ExternalDocumentRef:DocumentRef-spdx-tool-2.1 http://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3-9A0C-0305E82C3301 SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759
        '''
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'EXT_DOC_REF',
                                 'ExternalDocumentRef', 2)
        self.token_assert_helper(self.l.token(), 'DOC_REF_ID',
                                 'DocumentRef-spdx-tool-2.1', 2)
        self.token_assert_helper(self.l.token(), 'DOC_URI',
                                 'http://spdx.org/spdxdocs/spdx-tools-v2.1-3F25'
                                 '04E0-4F89-41D3-9A0C-0305E82C3301', 2)
        self.token_assert_helper(self.l.token(), 'EXT_DOC_REF_CHKSUM',
                                 'SHA1: '
                                 'd6a770ba38583ed4bb4525bd96e50461655d2759', 2)

    def test_creation_info(self):
        data = creation_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'CREATOR', 'Creator', 1)
        self.token_assert_helper(self.l.token(), 'PERSON_VALUE', "Person: Bob (bob@example.com)", 1)
        self.token_assert_helper(self.l.token(), 'CREATOR', 'Creator', 2)
        self.token_assert_helper(self.l.token(), 'ORG_VALUE', 'Organization: Acme.', 2)
        self.token_assert_helper(self.l.token(), 'CREATED', 'Created', 3)
        self.token_assert_helper(self.l.token(), 'DATE', '2010-02-03T00:00:00Z', 3)
        self.token_assert_helper(self.l.token(), 'CREATOR_COMMENT', 'CreatorComment', 4)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Sample Comment</text>', 4)

    def test_review_info(self):
        data = review_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'REVIEWER', 'Reviewer', 1)
        self.token_assert_helper(self.l.token(), 'PERSON_VALUE', "Person: Bob the Reviewer", 1)
        self.token_assert_helper(self.l.token(), 'REVIEW_DATE', 'ReviewDate', 2)
        self.token_assert_helper(self.l.token(), 'DATE', '2010-02-10T00:00:00Z', 2)
        self.token_assert_helper(self.l.token(), 'REVIEW_COMMENT', 'ReviewComment', 3)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Bob was Here.</text>', 3)
        self.token_assert_helper(self.l.token(), 'REVIEWER', 'Reviewer', 4)
        self.token_assert_helper(self.l.token(), 'PERSON_VALUE', "Person: Alice the Reviewer", 4)
        self.token_assert_helper(self.l.token(), 'REVIEW_DATE', 'ReviewDate', 5)
        self.token_assert_helper(self.l.token(), 'DATE', '2011-02-10T00:00:00Z', 5)
        self.token_assert_helper(self.l.token(), 'REVIEW_COMMENT', 'ReviewComment', 6)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Alice was also here.</text>', 6)

    def test_package(self):
        data = package_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'PKG_NAME', 'PackageName', 1)
        self.token_assert_helper(self.l.token(), 'LINE', 'Test', 1)
        self.token_assert_helper(self.l.token(), 'SPDX_ID', 'SPDXID', 2)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDXRef-Package', 2)
        self.token_assert_helper(self.l.token(), 'PKG_VERSION', 'PackageVersion', 3)
        self.token_assert_helper(self.l.token(), 'LINE', '1:22.36.1-8+deb11u1', 3)
        self.token_assert_helper(self.l.token(), 'PKG_DOWN', 'PackageDownloadLocation', 4)
        self.token_assert_helper(self.l.token(), 'LINE', 'http://example.com/test', 4)
        self.token_assert_helper(self.l.token(), 'PKG_FILES_ANALYZED', 'FilesAnalyzed', 5)
        self.token_assert_helper(self.l.token(), 'LINE', 'True', 5)
        self.token_assert_helper(self.l.token(), 'PKG_SUM', 'PackageSummary', 6)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Test package</text>', 6)
        self.token_assert_helper(self.l.token(), 'PKG_SRC_INFO', 'PackageSourceInfo', 7)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Version 1.0 of test</text>', 7)
        self.token_assert_helper(self.l.token(), 'PKG_FILE_NAME', 'PackageFileName', 8)
        self.token_assert_helper(self.l.token(), 'LINE', 'test-1.0.zip', 8)
        self.token_assert_helper(self.l.token(), 'PKG_SUPPL', 'PackageSupplier', 9)
        self.token_assert_helper(self.l.token(), 'ORG_VALUE', 'Organization:ACME', 9)
        self.token_assert_helper(self.l.token(), 'PKG_ORIG', 'PackageOriginator', 10)
        self.token_assert_helper(self.l.token(), 'ORG_VALUE', 'Organization:ACME', 10)
        self.token_assert_helper(self.l.token(), 'PKG_CHKSUM', 'PackageChecksum', 11)
        self.token_assert_helper(self.l.token(), 'CHKSUM', 'SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12', 11)
        self.token_assert_helper(self.l.token(), 'PKG_VERF_CODE', 'PackageVerificationCode', 12)
        self.token_assert_helper(self.l.token(), 'LINE', '4e3211c67a2d28fced849ee1bb76e7391b93feba (something.rdf, something.txt)', 12)
        self.token_assert_helper(self.l.token(), 'PKG_DESC', 'PackageDescription', 13)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>A package.</text>', 13)
        self.token_assert_helper(self.l.token(), 'PKG_COMMENT', 'PackageComment', 14)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Comment on the package.</text>', 14)
        self.token_assert_helper(self.l.token(), 'PKG_CPY_TEXT', 'PackageCopyrightText', 15)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text> Copyright 2014 Acme Inc.</text>', 15)
        self.token_assert_helper(self.l.token(), 'PKG_LICS_DECL', 'PackageLicenseDeclared', 16)
        self.token_assert_helper(self.l.token(), 'LINE', 'Apache-2.0', 16)
        self.token_assert_helper(self.l.token(), 'PKG_LICS_CONC', 'PackageLicenseConcluded', 17)
        self.token_assert_helper(self.l.token(), 'LINE', '(LicenseRef-2.0 and Apache-2.0)', 17)
        self.token_assert_helper(self.l.token(), 'PKG_LICS_FFILE', 'PackageLicenseInfoFromFiles', 18)
        self.token_assert_helper(self.l.token(), 'LINE', 'Apache-1.0', 18)
        self.token_assert_helper(self.l.token(), 'PKG_LICS_FFILE', 'PackageLicenseInfoFromFiles', 19)
        self.token_assert_helper(self.l.token(), 'LINE', 'Apache-2.0', 19)
        self.token_assert_helper(self.l.token(), 'PKG_LICS_COMMENT', 'PackageLicenseComments', 20)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>License Comments</text>', 20)
        self.token_assert_helper(self.l.token(), 'PKG_EXT_REF', 'ExternalRef', 21)
        self.token_assert_helper(self.l.token(), 'LINE', 'SECURITY cpe23Type cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:', 21)
        self.token_assert_helper(self.l.token(), 'PKG_EXT_REF_COMMENT', 'ExternalRefComment', 22)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Some comment about the package.</text>', 22)
        self.token_assert_helper(self.l.token(), 'PRIMARY_PACKAGE_PURPOSE', 'PrimaryPackagePurpose', 23)
        self.token_assert_helper(self.l.token(), 'OPERATING_SYSTEM', 'OPERATING-SYSTEM', 23)
        self.token_assert_helper(self.l.token(), 'BUILT_DATE', 'BuiltDate', 24)
        self.token_assert_helper(self.l.token(), 'DATE', '2020-01-01T12:00:00Z', 24)
        self.token_assert_helper(self.l.token(), 'RELEASE_DATE', 'ReleaseDate', 25)
        self.token_assert_helper(self.l.token(), 'DATE', '2021-01-01T12:00:00Z', 25)
        self.token_assert_helper(self.l.token(), 'VALID_UNTIL_DATE', 'ValidUntilDate', 26)
        self.token_assert_helper(self.l.token(), 'DATE', '2022-01-01T12:00:00Z', 26)


    def test_unknown_tag(self):
        data = unknown_tag_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'UNKNOWN_TAG', 'SomeUnknownTag', 1)
        self.token_assert_helper(self.l.token(), 'LINE', 'SomeUnknownValue', 1)

    def test_snippet(self):
        data = snippet_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'SNIPPET_SPDX_ID', 'SnippetSPDXID', 1)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDXRef-Snippet', 1)
        self.token_assert_helper(self.l.token(), 'SNIPPET_LICS_COMMENT', 'SnippetLicenseComments', 2)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Some lic comment.</text>', 2)
        self.token_assert_helper(self.l.token(), 'SNIPPET_CR_TEXT', 'SnippetCopyrightText', 3)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text> Copyright 2008-2010 John Smith </text>', 3)
        self.token_assert_helper(self.l.token(), 'SNIPPET_COMMENT', 'SnippetComment', 4)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Some snippet comment.</text>', 4)
        self.token_assert_helper(self.l.token(), 'SNIPPET_NAME', 'SnippetName', 5)
        self.token_assert_helper(self.l.token(), 'LINE', 'from linux kernel', 5)
        self.token_assert_helper(self.l.token(), 'SNIPPET_FILE_SPDXID',
                                 'SnippetFromFileSPDXID', 6)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDXRef-DoapSource', 6)
        self.token_assert_helper(self.l.token(), 'SNIPPET_LICS_CONC',
                                 'SnippetLicenseConcluded', 7)
        self.token_assert_helper(self.l.token(), 'LINE', 'Apache-2.0', 7)
        self.token_assert_helper(self.l.token(), 'SNIPPET_LICS_INFO',
                                 'LicenseInfoInSnippet', 8)
        self.token_assert_helper(self.l.token(), 'LINE', 'Apache-2.0', 8)
        self.token_assert_helper(self.l.token(), 'SNIPPET_BYTE_RANGE', 'SnippetByteRange', 9)
        self.token_assert_helper(self.l.token(), 'LINE', '310:420', 9)
        self.token_assert_helper(self.l.token(), 'SNIPPET_LINE_RANGE', 'SnippetLineRange', 10)
        self.token_assert_helper(self.l.token(), 'LINE', '5:7', 10)

    def test_annotation(self):
        data = annotation_str
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'ANNOTATOR', 'Annotator', 1)
        self.token_assert_helper(self.l.token(), 'PERSON_VALUE', 'Person: Jane Doe()', 1)
        self.token_assert_helper(self.l.token(), 'ANNOTATION_DATE', 'AnnotationDate', 2)
        self.token_assert_helper(self.l.token(), 'DATE', '2010-01-29T18:30:22Z', 2)
        self.token_assert_helper(self.l.token(), 'ANNOTATION_COMMENT', 'AnnotationComment', 3)
        self.token_assert_helper(self.l.token(), 'TEXT', '<text>Document level annotation</text>', 3)
        self.token_assert_helper(self.l.token(), 'ANNOTATION_TYPE', 'AnnotationType', 4)
        self.token_assert_helper(self.l.token(), 'OTHER', 'OTHER', 4)
        self.token_assert_helper(self.l.token(), 'ANNOTATION_SPDX_ID', 'SPDXREF', 5)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDXRef-DOCUMENT', 5)

    def test_correct_line_number_with_empty_line_between(self):
        data = document_str_with_empty_line
        self.l.input(data)
        self.token_assert_helper(self.l.token(), 'DOC_VERSION', 'SPDXVersion', 1)
        self.token_assert_helper(self.l.token(), 'LINE', 'SPDX-2.1', 1)
        self.token_assert_helper(self.l.token(), 'DOC_LICENSE', 'DataLicense', 3)
        self.token_assert_helper(self.l.token(), 'LINE', 'CC0-1.0', 3)

    def token_assert_helper(self, token, ttype, value, line):
        assert token.type == ttype
        assert token.value == value
        assert token.lineno == line


class TestParser(TestCase):
    maxDiff = None
    complete_str = '{0}\n{1}\n{2}\n{3}\n{4}\n{5}\n{6}'.format(document_str, creation_str, review_str, package_str,
                                                              file_str, annotation_str, snippet_str)

    def setUp(self):
        self.p = Parser(Builder(), StandardLogger())
        self.p.build()

    def test_doc(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert document.version == Version(major=2, minor=1)
        assert document.data_license.identifier == 'CC0-1.0'
        assert document.name == 'Sample_Document-V2.1'
        assert document.spdx_id == 'SPDXRef-DOCUMENT'
        assert document.comment == 'Sample Comment'
        assert document.namespace == 'https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301'

    def test_creation_info(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert len(document.creation_info.creators) == 2
        assert document.creation_info.comment == 'Sample Comment'
        assert (document.creation_info.created_iso_format == '2010-02-03T00:00:00Z')

    def test_review(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert len(document.reviews) == 2

    def test_package(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert document.package.name == 'Test'
        assert document.package.spdx_id == 'SPDXRef-Package'
        assert document.package.version == '1:22.36.1-8+deb11u1'
        assert len(document.package.licenses_from_files) == 2
        assert (document.package.conc_lics.identifier == 'LicenseRef-2.0 AND Apache-2.0')
        assert document.package.files_analyzed is True
        assert document.package.comment == 'Comment on the package.'
        assert document.package.pkg_ext_refs[-1].category == 'SECURITY'
        assert document.package.pkg_ext_refs[-1].pkg_ext_ref_type == 'cpe23Type'
        assert document.package.pkg_ext_refs[-1].locator == 'cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:'
        assert document.package.pkg_ext_refs[-1].comment == 'Some comment about the package.'
        assert document.package.primary_package_purpose == PackagePurpose.OPERATING_SYSTEM
        assert document.package.built_date == datetime(2020, 1, 1, 12, 0, 0)
        assert document.package.release_date == datetime(2021, 1, 1, 12, 0, 0)
        assert document.package.valid_until_date == datetime(2022, 1, 1, 12, 0, 0)

    def test_file(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert len(document.files) == 1
        spdx_file = document.files[0]
        assert spdx_file.name == 'testfile.java'
        assert spdx_file.spdx_id == 'SPDXRef-File'
        assert spdx_file.file_types == [spdx.file.FileType.SOURCE, spdx.file.FileType.TEXT]
        assert len(spdx_file.artifact_of_project_name) == 1
        assert len(spdx_file.artifact_of_project_home) == 1
        assert len(spdx_file.artifact_of_project_uri) == 1
        assert spdx_file.comment == 'Very long file'
        assert spdx_file.attribution_text == 'Acknowledgements that might be required to be communicated in ' \
                                             'some contexts.'

    def test_annotation(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert len(document.annotations) == 1
        assert document.annotations[-1].annotator.name == 'Jane Doe'
        assert spdx.utils.datetime_iso_format(document.annotations[-1].annotation_date) == '2010-01-29T18:30:22Z'
        assert document.annotations[-1].comment == 'Document level annotation'
        assert document.annotations[-1].annotation_type == 'OTHER'
        assert document.annotations[-1].spdx_id == 'SPDXRef-DOCUMENT'

    def test_unknown_tag(self):
        try:
            from StringIO import StringIO
        except ImportError:
            from io import StringIO

        saved_out = sys.stdout
        sys.stdout = StringIO()
        document, error = self.p.parse(unknown_tag_str)
        self.assertEqual(sys.stdout.getvalue(), 'Found unknown tag : SomeUnknownTag at line: 1\n')
        sys.stdout = saved_out
        assert error
        assert document is not None

    def test_snippet(self):
        document, error = self.p.parse(self.complete_str)
        assert document is not None
        assert not error
        assert len(document.snippet) == 1
        assert document.snippet[-1].spdx_id == 'SPDXRef-Snippet'
        assert document.snippet[-1].name == 'from linux kernel'
        assert document.snippet[-1].comment == 'Some snippet comment.'
        assert document.snippet[-1].copyright == ' Copyright 2008-2010 John Smith '
        assert document.snippet[-1].license_comment == 'Some lic comment.'
        assert document.snippet[-1].snip_from_file_spdxid == 'SPDXRef-DoapSource'
        assert document.snippet[-1].conc_lics.identifier == 'Apache-2.0'
        assert document.snippet[-1].licenses_in_snippet[-1].identifier == 'Apache-2.0'
        assert document.snippet[-1].byte_range[0] == 310
        assert document.snippet[-1].byte_range[1] == 420
        assert document.snippet[-1].line_range[0] == 5
        assert document.snippet[-1].line_range[1] == 7



# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import xmltodict

from spdx.parsers import jsonyamlxml


class Parser(jsonyamlxml.Parser):
    """
    Wrapper class for jsonyamlxml.Parser to provide an interface similar to
    RDF and TV Parser classes (i.e., spdx.parsers.<format name>.Parser) for XML parser.
    It also avoids to repeat jsonyamlxml.Parser.parse code for JSON, YAML and XML parsers
    """

    def __init__(self, builder, logger):
        super(Parser, self).__init__(builder, logger)
        self.LIST_LIKE_FIELDS = {
            "creators",
            "externalDocumentRefs",
            "extractedLicenseInfos",
            "seeAlsos",
            "annotations",
            "relationships",
            "snippets",
            "reviewers",
            "fileTypes",
            "licenseInfoFromFiles",
            "licenseInfoInFiles",
            "artifactOf",
            "fileContributors",
            "fileDependencies",
            "files",
            "documentDescribes",
            "packages",
            "checksums",
            "hasFiles",
            "externalRefs",
            "ranges",
            "licenseInfoInSnippets",
            "packageVerificationCodeExcludedFiles",
        }

    def parse(self, file):
        parsed_xml = xmltodict.parse(
            file.read(), strip_whitespace=False, encoding="utf-8"
        )
        fixed_object = self._set_in_list(parsed_xml, self.LIST_LIKE_FIELDS)
        self.document_object = fixed_object.get("Document")
        return super(Parser, self).parse()

    def _set_in_list(self, data, keys):
        """
        xmltodict parse list-like fields in different way when there is only one
        of them than when there are several of them.
        Set in lists those fields that are expected to be in them.
        """
        if isinstance(data, (dict, OrderedDict)):
            new_data = OrderedDict()
            for k, v in data.items():
                if k in keys and not isinstance(v, list):
                    new_data[k] = [self._set_in_list(v, keys)]
                else:
                    new_data[k] = self._set_in_list(v, keys)
            return new_data
        elif isinstance(data, list):
            new_data = []
            for element in data:
                new_data.append(self._set_in_list(element, keys))
            return new_data
        return data



# Copyright (c) 2020 Yash Varshney
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import auto, Enum

from spdx.parsers.loggers import ErrorMessages


class RelationshipType(Enum):
    AMENDS = auto()
    OTHER = auto()
    COPY_OF = auto()
    TEST_OF = auto()
    ANCESTOR_OF = auto()
    BUILD_DEPENDENCY_OF = auto()
    BUILD_TOOL_OF = auto()
    CONTAINED_BY = auto()
    CONTAINS = auto()
    DATA_FILE_OF = auto()
    DEPENDENCY_MANIFEST_OF = auto()
    DEPENDENCY_OF = auto()
    DEPENDS_ON = auto()
    DESCENDANT_OF = auto()
    DESCRIBED_BY = auto()
    DESCRIBES = auto()
    DEV_DEPENDENCY_OF = auto()
    DEV_TOOL_OF = auto()
    DISTRIBUTION_ARTIFACT = auto()
    DOCUMENTATION_OF = auto()
    DYNAMIC_LINK = auto()
    EXAMPLE_OF = auto()
    EXPANDED_FROM_ARCHIVE = auto()
    FILE_ADDED = auto()
    FILE_DELETED = auto()
    FILE_MODIFIED = auto()
    GENERATED_FROM = auto()
    GENERATES = auto()
    HAS_PREREQUISITE = auto()
    METAFILE_OF = auto()
    OPTIONAL_COMPONENT_OF = auto()
    OPTIONAL_DEPENDENCY_OF = auto()
    PACKAGE_OF = auto()
    PATCH_APPLIED = auto()
    PATCH_FOR = auto()
    PREREQUISITE_FOR = auto()
    PROVIDED_DEPENDENCY_OF = auto()
    RUNTIME_DEPENDENCY_OF = auto()
    STATIC_LINK = auto()
    TEST_CASE_OF = auto()
    TEST_DEPENDENCY_OF = auto()
    TEST_TOOL_OF = auto()
    VARIANT_OF = auto()
    REQUIREMENT_DESCRIPTION_FOR = auto()
    SPECIFICATION_FOR = auto()


class Relationship(object):
    """
    Document relationship information
    Fields:
    - relationship:  provides information about the relationship between two SPDX elements.
    - relationship_comment:  place for the SPDX file creator to record any general comments. Optional, One
    """

    def __init__(self, relationship=None, relationship_comment=None):
        self.relationship = relationship
        self.relationship_comment = relationship_comment

    def __eq__(self, other):
        return (
            isinstance(other, Relationship)
            and self.relationship == other.relationship
        )

    @property
    def has_comment(self):
        return self.relationship_comment is not None

    @property
    def spdx_element_id(self):
        return self.relationship.split(" ")[0]

    @property
    def relationship_type(self):
        return self.relationship.split(" ")[1]

    @property
    def related_spdx_element(self):
        return self.relationship.split(" ")[2]

    def validate(self, messages: ErrorMessages) -> ErrorMessages:
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        r_type = self.relationship_type
        if r_type not in [name for name, _ in RelationshipType.__members__.items()]:
            messages.append(
                "Relationship type must be one of the constants defined in "
                "class spdx.relationship.Relationship"
            )
        return messages



# Copyright (c) 2021 spdx tool contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from spdx.parsers import parse_anything

dirname = os.path.join(os.path.dirname(__file__), "data", "formats")
# Some rdf examples still cause issues
test_files = [os.path.join(dirname, fn) for fn in os.listdir(dirname) if not "RdfExample-v" in fn]


@pytest.mark.parametrize("test_file", test_files)
def test_parse_anything(test_file):
    doc, error = parse_anything.parse_file(test_file)

    assert not error

    # test a few fields, the core of the tests are per parser
    assert doc.name in ('Sample_Document-V2.1', 'xyz-0.1.0', 'SPDX-Tools-v2.0')
    assert doc.comment in (None, 'This is a sample spreadsheet', 'Sample Comment',
                           'This document was created using SPDX 2.0 using licenses from the web site.')
    assert len(doc.packages) in (1, 2, 3, 4)



# Copyright (c) 2022 spdx tool contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import TestCase

from spdx.cli_tools.convertor import determine_infile_and_outfile

from tests.testing_utils import raises


class TestConvertor(TestCase):
    maxDiff = None

    def test_determine_input_with_known_i_o_format(self):
        infile_given = 'infile.rdf'
        outfile_given = 'outfile.json'
        src = ()
        from_ = None
        to = None

        infile, outfile = determine_infile_and_outfile(infile_given, outfile_given, src, from_, to)

        assert infile == infile_given
        assert outfile == outfile_given

    def test_determine_input_with_unknown_i_o_format(self):
        infile_given = None
        outfile_given = None
        src = ('infile.in', 'outfile.out')
        from_ = 'rdf'
        to = 'json'
        expected_infile = 'infile.rdf'
        expected_outfile = 'outfile.json'

        infile, outfile = determine_infile_and_outfile(infile_given, outfile_given, src, from_, to)

        assert infile == expected_infile
        assert outfile == expected_outfile

    def test_determine_input_with_known_i_format_unknown_o_format(self):
        infile_given = 'infile.rdf'
        outfile_given = None
        src = ('outfile',)
        from_ = None
        to = 'json'
        expected_outfile = 'outfile.json'

        infile, outfile = determine_infile_and_outfile(infile_given, outfile_given, src, from_, to)

        assert infile == infile_given
        assert outfile == expected_outfile

    def test_determine_input_with_unknown_i_format_known_o_format(self):
        infile_given = None
        outfile_given = 'outfile.json'
        src = ('infile',)
        from_ = 'rdf'
        to = None
        expected_infile = 'infile.rdf'

        infile, outfile = determine_infile_and_outfile(infile_given, outfile_given, src, from_, to)

        assert infile == expected_infile
        assert outfile == outfile_given

    @raises(ValueError)
    def test_determine_input_with_invalid_arguments(self):
        infile_given = None
        outfile_given = None
        src = ()
        from_ = None
        to = None

        infile, outfile = determine_infile_and_outfile(infile_given, outfile_given, src, from_, to)



import glob
import os
from datetime import datetime
from typing import List

import pytest

from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.document import Document
from spdx.file import File
from spdx.license import License
from spdx.package import Package, ExternalPackageRef, PackagePurpose
from spdx.parsers.parse_anything import parse_file
from spdx.relationship import Relationship
from spdx.snippet import Snippet
from spdx.utils import update_dict_item_with_new_item
from spdx.writers import write_anything
from tests.test_rdf_writer import minimal_document_with_package

tested_formats: List[str] = ['yaml', 'xml', 'json']


@pytest.fixture
def temporary_file_path() -> str:
    temporary_file_path = "temp_test_writer_output"
    yield temporary_file_path
    file_with_ending = glob.glob(temporary_file_path + "*")
    for file in file_with_ending:
        os.remove(file)


@pytest.mark.parametrize("out_format", tested_formats)
def test_external_package_references(temporary_file_path: str, out_format: str) -> None:
    document: Document = minimal_document_with_package()
    package: Package = document.packages[0]
    first_ref = ExternalPackageRef(category="PACKAGE-MANAGER")
    second_ref = ExternalPackageRef(category="SECURITY")
    package.add_pkg_ext_refs(first_ref)
    package.add_pkg_ext_refs(second_ref)

    file_path_with_ending = temporary_file_path + "." + out_format
    write_anything.write_file(document, file_path_with_ending, validate=False)

    parsed_document = parse_file(file_path_with_ending)[0]

    parsed_package: Package = parsed_document.packages[0]
    assert len(parsed_package.pkg_ext_refs) is 2
    written_reference_categories = list(map(lambda x: x.category, parsed_package.pkg_ext_refs))
    assert first_ref.category in written_reference_categories
    assert second_ref.category in written_reference_categories


@pytest.mark.parametrize("out_format", tested_formats)
def test_primary_package_purpose(temporary_file_path: str, out_format: str):
    document: Document = minimal_document_with_package()
    package: Package = document.packages[0]
    package.primary_package_purpose = PackagePurpose.OPERATING_SYSTEM

    file_path_with_ending = temporary_file_path + "." + out_format
    write_anything.write_file(document, file_path_with_ending, validate=False)

    parsed_document: Document = parse_file(file_path_with_ending)[0]
    parsed_package: Package = parsed_document.packages[0]

    assert parsed_package.primary_package_purpose == PackagePurpose.OPERATING_SYSTEM


@pytest.mark.parametrize("out_format", tested_formats)
def test_release_built_valid_until_date(temporary_file_path: str, out_format: str):
    document: Document = minimal_document_with_package()
    package: Package = document.packages[0]
    package.release_date = datetime(2021, 1, 1, 12, 0, 0)
    package.built_date = datetime(2021, 1, 1, 12, 0, 0)
    package.valid_until_date = datetime(2022, 1, 1, 12, 0, 0)

    file_path_with_ending = temporary_file_path + "." + out_format
    write_anything.write_file(document, file_path_with_ending, validate=False)

    parsed_document: Document = parse_file(file_path_with_ending)[0]
    parsed_package: Package = parsed_document.packages[0]

    assert parsed_package.release_date == package.release_date
    assert parsed_package.built_date == package.built_date
    assert parsed_package.valid_until_date == package.valid_until_date


@pytest.mark.parametrize("out_format", ['yaml', 'xml', 'json'])
def test_snippet_byte_range(temporary_file_path, out_format):
    snippet = minimal_snippet()
    parsed_snippet = write_and_parse_snippet(out_format, snippet, temporary_file_path)

    assert parsed_snippet.byte_range == (310, 420)
    assert parsed_snippet.line_range is None


@pytest.mark.parametrize("out_format", ['yaml', 'xml', 'json'])
def test_snippet_ranges(temporary_file_path, out_format):
    snippet = minimal_snippet()
    snippet.line_range = (5, 23)
    parsed_snippet = write_and_parse_snippet(out_format, snippet, temporary_file_path)

    assert parsed_snippet.byte_range == (310, 420)
    assert parsed_snippet.line_range == (5, 23)


def minimal_snippet():
    snippet = Snippet()
    snippet.spdx_id = "SPDXRef-Snippet"
    snippet.snip_from_file_spdxid = "SPDXRef-File"
    snippet.byte_range = (310, 420)
    return snippet


def write_and_parse_snippet(out_format, snippet, temporary_file_path):
    document: Document = minimal_document_with_package()
    document.add_snippet(snippet)
    file_path_with_ending = temporary_file_path + "." + out_format
    write_anything.write_file(document, file_path_with_ending, validate=False)
    parsed_document: Document = parse_file(file_path_with_ending)[0]
    parsed_snippet: Snippet = parsed_document.snippet[0]

    return parsed_snippet


@pytest.mark.parametrize("out_format", ['yaml', 'xml', 'json'])
def test_files_without_package(temporary_file_path, out_format):
    document: Document = minimal_document()
    document.spdx_id = "SPDXRef-DOCUMENT"
    file: File = minimal_file()
    document.add_file(file)
    describes_relationship: Relationship = Relationship("SPDXRef-DOCUMENT DESCRIBES SPDXRef-File")
    document.add_relationship(relationship=describes_relationship)

    file_path_with_ending = temporary_file_path + "." + out_format
    write_anything.write_file(document, file_path_with_ending, validate=False)

    parsed_document: Document = parse_file(file_path_with_ending)[0]
    parsed_file: File = parsed_document.files[0]

    assert parsed_file.name == "Example File"


def minimal_document():
    document = Document(data_license=License.from_identifier('CC0-1.0'))
    document.creation_info.set_created_now()
    return document


def minimal_file():
    file = File(name="Example File", spdx_id="SPDXRef-File")
    file.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'some-sha1-value'))
    return file


@pytest.mark.parametrize("key,value,expected_length",
                         [("existing_key", "new_value", 2), ("existing_key", "existing_value", 1),
                          ("new_key", "new_value", 1)])
def test_update_dict_item_with_new_item(key, value, expected_length):
    current_state = {"existing_key": ["existing_value"]}
    update_dict_item_with_new_item(current_state, key, value)

    assert key in current_state
    assert value in current_state[key]
    assert len(current_state[key]) == expected_length



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

import tests.testing_utils as testing_utils

from spdx.document import Document
from spdx.license import License
import spdx.parsers.tagvaluebuilders as builders
from spdx.version import Version


class TestDocumentBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.document = Document()
        self.builder = builders.DocBuilder()

    def test_correct_version(self):
        version_str = "SPDX-2.1"
        self.builder.set_doc_version(self.document, version_str)
        assert self.document.version.major == 2 and self.document.version.minor == 1

    @testing_utils.raises(builders.CardinalityError)
    def test_version_cardinality(self):
        version_str = "SPDX-2.1"
        self.builder.set_doc_version(self.document, version_str)
        self.builder.set_doc_version(self.document, version_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_version_value(self):
        version_str = "2.1"
        self.builder.set_doc_version(self.document, version_str)

    def test_correct_data_lics(self):
        lics_str = "CC0-1.0"
        self.builder.set_doc_data_lics(self.document, lics_str)
        assert self.document.data_license == License.from_identifier(lics_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_data_lics_value(self):
        lics_str = "GPL"
        self.builder.set_doc_data_lics(self.document, lics_str)

    @testing_utils.raises(builders.CardinalityError)
    def test_data_lics_cardinality(self):
        lics_str = "CC0-1.0"
        self.builder.set_doc_data_lics(self.document, lics_str)
        self.builder.set_doc_data_lics(self.document, lics_str)

    def test_correct_name(self):
        name_str = "Sample_Document-V2.1"
        self.builder.set_doc_name(self.document, name_str)
        assert self.document.name == name_str

    @testing_utils.raises(builders.CardinalityError)
    def test_name_cardinality(self):
        name_str = "Sample_Document-V2.1"
        self.builder.set_doc_name(self.document, name_str)
        self.builder.set_doc_name(self.document, name_str)

    def test_correct_doc_namespace(self):
        doc_namespace_str = "https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301"
        self.builder.set_doc_namespace(self.document, doc_namespace_str)
        assert self.document.namespace == doc_namespace_str

    @testing_utils.raises(builders.SPDXValueError)
    def test_doc_namespace_value(self):
        doc_namespace_str = "https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301#SPDXRef-DOCUMENT"
        self.builder.set_doc_data_lics(self.document, doc_namespace_str)

    @testing_utils.raises(builders.CardinalityError)
    def test_doc_namespace_cardinality(self):
        doc_namespace_str = "https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301"
        self.builder.set_doc_namespace(self.document, doc_namespace_str)
        self.builder.set_doc_namespace(self.document, doc_namespace_str)

    def test_correct_data_comment(self):
        comment_str = "This is a comment."
        comment_text = "<text>" + comment_str + "</text>"
        self.builder.set_doc_comment(self.document, comment_text)
        assert self.document.comment == comment_str

    @testing_utils.raises(builders.CardinalityError)
    def test_comment_cardinality(self):
        comment_str = "This is a comment."
        comment_text = "<text>" + comment_str + "</text>"
        self.builder.set_doc_comment(self.document, comment_text)
        self.builder.set_doc_comment(self.document, comment_text)

    @testing_utils.raises(builders.SPDXValueError)
    def test_comment_value(self):
        comment = "<text>sls\nlss<text"
        self.builder.set_doc_comment(self.document, comment)


class TestExternalDocumentRefBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.document = Document()
        self.builder = builders.ExternalDocumentRefBuilder()

    def test_external_doc_id(self):
        ext_doc_id = "DocumentRef-spdx-tool-2.1"
        self.builder.set_ext_doc_id(self.document, ext_doc_id)
        assert (
            self.document.ext_document_references[-1].external_document_id == ext_doc_id
        )

    def test_spdx_doc_uri(self):
        spdx_doc_uri = "https://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3-9A0C-0305E82C3301"
        self.builder.set_ext_doc_id(self.document, "DocumentRef-spdx-tool-2.1")
        self.builder.set_spdx_doc_uri(self.document, spdx_doc_uri)
        assert (
            self.document.ext_document_references[-1].spdx_document_uri == spdx_doc_uri
        )

    def test_checksum(self):
        chksum = "SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759"
        chksum_val = "d6a770ba38583ed4bb4525bd96e50461655d2759"
        self.builder.set_ext_doc_id(self.document, "DocumentRef-spdx-tool-2.1")
        self.builder.set_chksum(self.document, chksum)
        assert self.document.ext_document_references[-1].checksum.value == chksum_val

    def test_add_ext_doc_refs(self):
        ext_doc_id_val = "DocumentRef-spdx-tool-2.1"
        spdx_doc_uri = "http://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3-9A0C-0305E82C3301"
        chksum = "SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759"
        chksum_val = "d6a770ba38583ed4bb4525bd96e50461655d2759"

        self.builder.add_ext_doc_refs(
            self.document, ext_doc_id_val, spdx_doc_uri, chksum
        )
        assert (
            self.document.ext_document_references[-1].external_document_id
            == ext_doc_id_val
        )
        assert (
            self.document.ext_document_references[-1].spdx_document_uri == spdx_doc_uri
        )
        assert self.document.ext_document_references[-1].checksum.value == chksum_val


class TestEntityBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.builder = builders.EntityBuilder()
        self.document = Document()

    def test_tool(self):
        tool_name = "autoanal-2.0"
        tool_str = "Tool: " + tool_name
        tool = self.builder.build_tool(self.document, tool_str)
        assert tool.name == tool_name

    @testing_utils.raises(builders.SPDXValueError)
    def test_tool_value_error(self):
        tool_str = "tool: ll"
        self.builder.build_tool(self.document, tool_str)

    def test_org_with_email(self):
        org_name = "Example"
        org_email = "example@example.org"
        org_str = "Organization: {0} ( {1} )".format(org_name, org_email)
        org = self.builder.build_org(self.document, org_str)
        assert org.name == org_name
        assert org.email == org_email

    def test_org(self):
        org_name = "Example"
        org_str = "Organization: {0} ()".format(org_name)
        org = self.builder.build_org(self.document, org_str)
        assert org.name == org_name
        assert org.email is None

    @testing_utils.raises(builders.SPDXValueError)
    def test_org_value_error(self):
        org_name = "Example"
        org_str = "Organization {0}".format(org_name)
        self.builder.build_org(self.document, org_str)

    def test_person_with_email(self):
        per_name = "Bob"
        per_email = "bob@example.org"
        per_str = "Person: {0} ( {1} )".format(per_name, per_email)
        per = self.builder.build_person(self.document, per_str)
        assert per.name == per_name
        assert per.email == per_email

    def test_per(self):
        per_name = "Bob"
        per_str = "Person: {0} ()".format(per_name)
        per = self.builder.build_person(self.document, per_str)
        assert per.name == per_name
        assert per.email is None

    @testing_utils.raises(builders.SPDXValueError)
    def test_per_value_error(self):
        per_name = "Bob"
        per_str = "Person {0}".format(per_name)
        self.builder.build_person(self.document, per_str)


class TestCreationInfoBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.document = Document()
        self.builder = builders.CreationInfoBuilder()
        self.entity_builder = builders.EntityBuilder()

    def test_add_creator(self):
        per_str = "Person: Bob (bob@example.com)"
        per = self.entity_builder.build_person(self.document, per_str)
        assert self.builder.add_creator(self.document, per)
        assert len(self.document.creation_info.creators) == 1
        assert self.document.creation_info.creators[0] == per

    @testing_utils.raises(builders.SPDXValueError)
    def test_invalid_creator_type(self):
        self.builder.add_creator(self.document, "hello")

    def test_created(self):
        created_str = "2010-02-03T00:00:00Z"
        assert self.builder.set_created_date(self.document, created_str)

    @testing_utils.raises(builders.CardinalityError)
    def test_more_than_one_created(self):
        created_str = "2010-02-03T00:00:00Z"
        self.builder.set_created_date(self.document, created_str)
        self.builder.set_created_date(self.document, created_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_created_value(self):
        created_str = "2010-02-03T00:00:00"
        self.builder.set_created_date(self.document, created_str)

    def test_license_list_vers(self):
        vers_str = "1.2"
        assert self.builder.set_lics_list_ver(self.document, vers_str)
        assert self.document.creation_info.license_list_version == Version(1, 2)

    @testing_utils.raises(builders.SPDXValueError)
    def test_lics_list_ver_value(self):
        self.builder.set_lics_list_ver(self.document, "1 2")

    @testing_utils.raises(builders.CardinalityError)
    def test_lics_list_ver_card(self):
        self.builder.set_lics_list_ver(self.document, "1.2")
        self.builder.set_lics_list_ver(self.document, "1.3")


class TestReviewBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.entity_builder = builders.EntityBuilder()
        self.builder = builders.ReviewBuilder()
        self.document = Document()

    @testing_utils.raises(builders.OrderError)
    def test_reviewed_without_reviewer(self):
        date_str = "2010-02-03T00:00:00Z"
        self.builder.add_review_date(self.document, date_str)

    @testing_utils.raises(builders.OrderError)
    def test_comment_without_reviewer(self):
        comment = "<text>Comment</text>"
        self.builder.add_review_comment(self.document, comment)

    @testing_utils.raises(builders.CardinalityError)
    def test_comment_cardinality(self):
        comment = "<text>Comment</text>"
        self.add_reviewer()
        assert self.builder.add_review_comment(self.document, comment)
        self.builder.add_review_comment(self.document, comment)

    @testing_utils.raises(builders.CardinalityError)
    def test_reviewed_cardinality(self):
        date_str = "2010-02-03T00:00:00Z"
        self.add_reviewer()
        assert self.builder.add_review_date(self.document, date_str)
        self.builder.add_review_date(self.document, date_str)

    def test_comment_reset(self):
        comment = "<text>Comment</text>"
        self.add_reviewer()
        assert self.builder.add_review_comment(self.document, comment)
        self.add_reviewer()
        assert self.builder.add_review_comment(self.document, comment)

    def test_reviewed_reset(self):
        date_str = "2010-02-03T00:00:00Z"
        self.add_reviewer()
        assert self.builder.add_review_date(self.document, date_str)
        self.add_reviewer()
        assert self.builder.add_review_date(self.document, date_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_date_value(self):
        date_str = "2010-2-03T00:00:00Z"
        self.add_reviewer()
        self.builder.add_review_date(self.document, date_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_comment_value(self):
        comment = "<text>Multiline \nComment<text>"
        self.add_reviewer()
        self.builder.add_review_comment(self.document, comment)

    def add_reviewer(self):
        per_str = "Person: Bob (bob@example.com)"
        per = self.entity_builder.build_person(self.document, per_str)
        self.builder.add_reviewer(self.document, per)


class TestAnnotationBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.entity_builder = builders.EntityBuilder()
        self.builder = builders.AnnotationBuilder()
        self.document = Document()

    @testing_utils.raises(builders.OrderError)
    def test_annotation_without_annotator(self):
        date_str = "2014-08-06T00:00:00Z"
        self.builder.add_annotation_date(self.document, date_str)

    @testing_utils.raises(builders.OrderError)
    def test_comment_without_annotator(self):
        comment = "<text>Comment without annotator</text>"
        self.builder.add_annotation_comment(self.document, comment)

    @testing_utils.raises(builders.OrderError)
    def test_type_without_annotator(self):
        annotation_type = "REVIEW"
        self.builder.add_annotation_type(self.document, annotation_type)

    @testing_utils.raises(builders.OrderError)
    def test_spdx_id_without_annotator(self):
        spdx_id = "SPDXRef-45"
        self.builder.set_annotation_spdx_id(self.document, spdx_id)

    @testing_utils.raises(builders.CardinalityError)
    def test_annotation_comment_cardinality(self):
        comment = "<text>Annotation Comment</text>"
        self.add_annotator()
        assert self.builder.add_annotation_comment(self.document, comment)
        self.builder.add_annotation_comment(self.document, comment)

    @testing_utils.raises(builders.CardinalityError)
    def test_annotation_cardinality(self):
        date_str = "2014-08-06T00:00:00Z"
        self.add_annotator()
        assert self.builder.add_annotation_date(self.document, date_str)
        self.builder.add_annotation_date(self.document, date_str)

    @testing_utils.raises(builders.CardinalityError)
    def test_annotation_spdx_id_cardinality(self):
        spdx_id = "SPDXRef-45"
        self.add_annotator()
        self.builder.set_annotation_spdx_id(self.document, spdx_id)
        self.builder.set_annotation_spdx_id(self.document, spdx_id)

    def test_annotation_comment_reset(self):
        comment = "<text>Annotation Comment</text>"
        self.add_annotator()
        assert self.builder.add_annotation_comment(self.document, comment)
        self.add_annotator()
        assert self.builder.add_annotation_comment(self.document, comment)

    def test_annotation_reset(self):
        date_str = "2014-08-06T00:00:00Z"
        self.add_annotator()
        assert self.builder.add_annotation_date(self.document, date_str)
        self.add_annotator()
        assert self.builder.add_annotation_date(self.document, date_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_annotation_date_value(self):
        date_str = "2014-8-06T00:00:00Z"
        self.add_annotator()
        self.builder.add_annotation_date(self.document, date_str)

    @testing_utils.raises(builders.SPDXValueError)
    def test_annotation_comment_value(self):
        comment = "<text>Annotation Comment\nthat spans multiple lines<text>"
        self.add_annotator()
        self.builder.add_annotation_comment(self.document, comment)

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_annotation_type_value(self):
        annotation_type = "Some random value instead of REVIEW or OTHER"
        self.add_annotator()
        self.builder.add_annotation_type(self.document, annotation_type)

    def test_correct_annotation_type(self):
        annotation_type = "REVIEW"
        self.add_annotator()
        assert self.builder.add_annotation_type(self.document, annotation_type)

    def test_correct_annotation_spdx_id(self):
        spdx_id = "SPDXRef-45"
        self.add_annotator()
        self.builder.set_annotation_spdx_id(self.document, spdx_id)

    @testing_utils.raises(builders.CardinalityError)
    def test_annotation_type_cardinality(self):
        annotation_type = "REVIEW"
        self.add_annotator()
        assert self.builder.add_annotation_type(self.document, annotation_type)
        self.builder.add_annotation_type(self.document, annotation_type)

    def add_annotator(self):
        per_str = "Person: Jim (jim@example.com)"
        per = self.entity_builder.build_person(self.document, per_str)
        self.builder.add_annotator(self.document, per)


class TestRelationshipBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.entity_builder = builders.EntityBuilder()
        self.builder = builders.RelationshipBuilder()
        self.document = Document()

    @testing_utils.raises(builders.OrderError)
    def test_relationship_comment_without_relationship(self):
        comment = "<text>Comment without relationship</text>"
        self.builder.add_relationship_comment(self.document, comment)

    def test_relationship_comment_reset(self):
        comment = "<text>Relationship Comment</text>"
        self.add_relationship()
        assert self.builder.add_relationship_comment(self.document, comment)
        self.add_relationship()
        assert self.builder.add_relationship_comment(self.document, comment)

    @testing_utils.raises(builders.SPDXValueError)
    def test_relationship_comment_value(self):
        comment = "<text>Relationship \n" \
                  "Comment<text>"
        self.add_relationship()
        self.builder.add_relationship_comment(self.document, comment)

    def test_correct_relationship(self):
        relationship = "SPDXRef-DOCUMENT DESCRIBES SPDXRef-File"
        assert self.builder.add_relationship(self.document, relationship)

    def add_relationship(self):
        relate_str = "SPDXRef-DOCUMENT DESCRIBES SPDXRef-File"
        self.builder.add_relationship(self.document, relate_str)


class TestPackageBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.builder = builders.PackageBuilder()
        self.document = Document()
        self.entity_builder = builders.EntityBuilder()

    def test_package_cardinality(self):
        assert self.builder.create_package(self.document, "pkg1")
        self.builder.create_package(self.document, "pkg2")

    def make_package(self):
        self.builder.create_package(self.document, "pkg")

    def make_person(self):
        per_str = "Person: Bob (bob@example.com)"
        per = self.entity_builder.build_person(self.document, per_str)
        return per

    @testing_utils.raises(builders.OrderError)
    def test_vers_order(self):
        self.builder.set_pkg_vers(self.document, "1.1")

    @testing_utils.raises(builders.OrderError)
    def test_file_name_order(self):
        self.builder.set_pkg_file_name(self.document, "test.jar")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_supplier_order(self):
        self.builder.set_pkg_supplier(self.document, self.make_person())

    @testing_utils.raises(builders.OrderError)
    def test_pkg_originator_order(self):
        self.builder.set_pkg_originator(self.document, self.make_person())

    @testing_utils.raises(builders.OrderError)
    def test_pkg_down_loc_order(self):
        self.builder.set_pkg_down_location(self.document, "http://example.com/pkg")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_home_order(self):
        self.builder.set_pkg_home(self.document, "http://example.com")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_verif_order(self):
        self.builder.set_pkg_verif_code(self.document, "some code")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_chksum_order(self):
        self.builder.set_pkg_checksum(self.document, "some code")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_source_info_order(self):
        self.builder.set_pkg_source_info(self.document, "hello")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_licenses_concluded_order(self):
        self.builder.set_pkg_licenses_concluded(self.document, "some license")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_lics_from_file_order(self):
        self.builder.set_pkg_license_from_file(self.document, "some license")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_lics_decl_order(self):
        self.builder.set_pkg_license_declared(self.document, "license")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_lics_comment_order(self):
        self.builder.set_pkg_license_comment(self.document, "<text>hello</text>")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_attribution_text_order(self):
        self.builder.set_pkg_attribution_text(self.document, "<text>hello</text>")

    def test_correct_pkg_attribution_text(self):
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_attribution_text(self.document, "<text>something</text>")

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_pkg_attribution_text(self):
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_attribution_text(self.document, "<tex> text over multiple lines\n with wrong tag </text>")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_cr_text_order(self):
        self.builder.set_pkg_cr_text(self.document, "<text>Something</text>")

    @testing_utils.raises(builders.OrderError)
    def test_pkg_summary_order(self):
        self.builder.set_pkg_summary(self.document, "<text>Something</text>")

    @testing_utils.raises(builders.OrderError)
    def test_set_pkg_desc_order(self):
        self.builder.set_pkg_desc(self.document, "<text>something</text>")

    @testing_utils.raises(builders.OrderError)
    def test_set_pkg_spdx_id_order(self):
        self.builder.set_pkg_spdx_id(self.document, "SPDXRe-Package")

    @testing_utils.raises(builders.OrderError)
    def test_set_pkg_files_analyzed_order(self):
        self.builder.set_pkg_files_analyzed(self.document, "True")

    @testing_utils.raises(builders.OrderError)
    def test_set_pkg_comment_order(self):
        self.builder.set_pkg_comment(self.document, "<text>something</text>")

    def test_correct_pkg_comment(self):
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_comment(self.document, "<text>something</text>")

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_pkg_comment(self):
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_comment(self.document, "<text>text in multiple lines\n with wrong tag <text>")

    def test_correct_pkg_spdx_id(self):
        self.builder.create_package(self.document, "pkg")
        assert self.builder.set_pkg_spdx_id(self.document, "SPDXRef-Package")
        assert self.document.package.spdx_id == "SPDXRef-Package"

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_pkg_spdx_id(self):
        self.builder.create_package(self.document, "pkg")
        assert self.builder.set_pkg_spdx_id(self.document, "SPDXRe-Package")

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_pkg_files_analyzed(self):
        self.builder.create_package(self.document, "pkg")
        assert self.builder.set_pkg_files_analyzed(self.document, "XYZ")

    def test_correct_pkg_files_analyzed_1(self):
        self.builder.create_package(self.document, "pkg")
        assert self.builder.set_pkg_files_analyzed(self.document, "True")

    def test_correct_pkg_files_analyzed_2(self):
        self.builder.create_package(self.document, "pkg")
        assert self.builder.set_pkg_files_analyzed(self.document, "true")

    def test_correct_pkg_ext_ref_category(self):
        category = "SECURITY"
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_ext_ref_category(self.document, category)
        assert self.document.package.pkg_ext_refs[-1].category == category

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_pkg_ext_ref_category(self):
        category = "some_other_value"
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_ext_ref_category(self.document, category)

    def test_correct_pkg_ext_ref_type(self):
        pkg_ext_ref_type = "cpe23Type"
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_ext_ref_type(self.document, pkg_ext_ref_type)
        assert (
            self.document.package.pkg_ext_refs[-1].pkg_ext_ref_type == pkg_ext_ref_type
        )

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_pkg_ext_ref_type(self):
        pkg_ext_ref_type = "cpe23Type with whitespace"
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_ext_ref_type(self.document, pkg_ext_ref_type)

    def test_correct_pkg_ext_ref_locator(self):
        locator = "cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*"
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_ext_ref_locator(self.document, locator)
        assert self.document.package.pkg_ext_refs[-1].locator == locator

    @testing_utils.raises(builders.OrderError)
    def test_pkg_ext_ref_without_pkg(self):
        locator = "cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*"
        self.builder.set_pkg_ext_ref_locator(self.document, locator)

    def test_correct_pkg_ext_comment(self):
        comment_str = "This is a comment."
        comment_text = "<text>" + comment_str + "</text>"
        self.builder.create_package(self.document, "pkg")
        self.builder.set_pkg_ext_ref_category(self.document, "SECURITY")
        self.builder.add_pkg_ext_ref_comment(self.document, comment_text)
        assert self.document.package.pkg_ext_refs[-1].comment == comment_str

    @testing_utils.raises(builders.OrderError)
    def test_pkg_ext_comment_without_pkg_ext_ref(self):
        comment_str = "This is a comment."
        comment_text = "<text>" + comment_str + "</text>"
        self.builder.create_package(self.document, "pkg")
        self.builder.add_pkg_ext_ref_comment(self.document, comment_text)


class TestSnippetBuilder(TestCase):
    maxDiff = None

    def setUp(self):
        self.entity_builder = builders.EntityBuilder()
        self.builder = builders.SnippetBuilder()
        self.document = Document()

    def test_create_snippet(self):
        assert self.builder.create_snippet(self.document, "SPDXRef-Snippet")

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_snippet_spdx_id(self):
        self.builder.create_snippet(self.document, "Some_value_with_$%")

    def test_snippet_name(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_name(self.document, "Name_of_snippet")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_name_order(self):
        self.builder.set_snippet_name(self.document, "Name_of_snippet")

    def test_snippet_comment(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_comment(self.document, "<text>Comment</text>")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_comment_order(self):
        self.builder.set_snippet_comment(self.document, "<text>Comment</text>")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_comment_text_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_comment(self.document, "Comment over multiple lines\n without enclosing tags.")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_attribution_text_order(self):
        self.builder.set_snippet_attribution_text(self.document, "<text>hello</text>")

    def test_correct_snippet_attribution_text(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_attribution_text(
            self.document, "<text>something</text>"
        )

    @testing_utils.raises(builders.SPDXValueError)
    def test_incorrect_snippet_attribution_text(self):
        self.builder.create_snippet(self.document, "SPDXRef-Package")
        self.builder.set_snippet_attribution_text(self.document, "Attribution text over multiple lines"
                                                                 "\n without enclosing tags.")

    def test_snippet_copyright(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_copyright(
            self.document, "<text>Copyright 2008-2010 John Smith</text>"
        )

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_copyright_text_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_copyright(
            self.document, "Copyright 2008-2010 John Smith\n over multiple lines without enclosing tags."
        )

    @testing_utils.raises(builders.OrderError)
    def test_snippet_copyright_order(self):
        self.builder.set_snippet_copyright(
            self.document, "<text>Copyright 2008-2010 John Smith</text>"
        )

    def test_snippet_lic_comment(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_lic_comment(self.document, "<text>Lic comment</text>")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_lic_comment_text_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_lic_comment(self.document, "Lic comment over multiple lines\n without tags")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_lic_comment_order(self):
        self.builder.set_snippet_lic_comment(self.document, "<text>Lic comment</text>")

    def test_snippet_from_file_spdxid(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snip_from_file_spdxid(self.document, "SPDXRef-DoapSource")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_from_file_spdxid_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snip_from_file_spdxid(self.document, "#_$random_chars")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_from_file_spdxid_order(self):
        self.builder.set_snip_from_file_spdxid(self.document, "SPDXRef-DoapSource")

    @testing_utils.raises(builders.CardinalityError)
    def test_snippet_from_file_spdxid_cardinality(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snip_from_file_spdxid(self.document, "SPDXRef-DoapSource")
        self.builder.set_snip_from_file_spdxid(self.document, "SPDXRef-somevalue")

    def test_snippet_conc_lics(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snip_concluded_license(
            self.document, License.from_identifier("Apache-2.0")
        )

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_conc_lics_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snip_concluded_license(self.document, "Apache-2.0")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_conc_lics_order(self):
        self.builder.set_snip_concluded_license(
            self.document, License.from_identifier("Apache-2.0")
        )

    @testing_utils.raises(builders.CardinalityError)
    def test_snippet_conc_lics_cardinality(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snip_concluded_license(
            self.document, License.from_identifier("Apache-2.0")
        )
        self.builder.set_snip_concluded_license(
            self.document, License.from_identifier("Apache-2.0")
        )

    def test_snippet_lics_info(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_lics_info(
            self.document, License.from_identifier("Apache-2.0")
        )
        self.builder.set_snippet_lics_info(
            self.document, License.from_identifier("GPL-2.0-or-later")
        )

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_lics_info_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_lics_info(self.document, "Apache-2.0")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_lics_info_order(self):
        self.builder.set_snippet_lics_info(
            self.document, License.from_identifier("Apache-2.0")
        )

    def test_snippet_byte_range(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_byte_range(self.document, "310:420")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_byte_range_order(self):
        self.builder.set_snippet_byte_range(self.document, "310:420")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_byte_range_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_byte_range(self.document, "310:30")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_byte_range_value_wrong_format(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_byte_range(self.document, "30")

    def test_snippet_line_range(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_line_range(self.document, "5:23")

    @testing_utils.raises(builders.OrderError)
    def test_snippet_line_range_order(self):
        self.builder.set_snippet_line_range(self.document, "5:23")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_line_range_value(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_line_range(self.document, "23:5")

    @testing_utils.raises(builders.SPDXValueError)
    def test_snippet_line_range_value_wrong_format(self):
        self.builder.create_snippet(self.document, "SPDXRef-Snippet")
        self.builder.set_snippet_line_range(self.document, "5:23d")



# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from spdx.parsers import jsonyamlxml


class Parser(jsonyamlxml.Parser):
    """
    Wrapper class for jsonyamlxml.Parser to provide an interface similar to
    RDF and TV Parser classes (i.e., spdx.parsers.<format name>.Parser) for YAML parser.
    It also avoids to repeat jsonyamlxml.Parser.parse code for JSON, YAML and XML parsers
    """

    def __init__(self, builder, logger):
        super(Parser, self).__init__(builder, logger)

    def parse(self, file):
        self.json_yaml_set_document(yaml.safe_load(file))
        return super(Parser, self).parse()



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Dict, Union

from spdx import file
from spdx import license
from spdx import package
from spdx import version
from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.document import Document
from spdx.parsers.builderexceptions import CardinalityError
from spdx.parsers.builderexceptions import OrderError
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.parsers import tagvaluebuilders
from spdx.parsers import validations
from spdx.parsers.rdf import convert_rdf_checksum_algorithm


class DocBuilder(object):
    VERS_STR_REGEX = re.compile(r"SPDX-(\d+)\.(\d+)", re.UNICODE)

    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_document()

    def set_doc_version(self, doc, value):
        """
        Set the document version.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        if not self.doc_version_set:
            self.doc_version_set = True
            m = self.VERS_STR_REGEX.match(value)
            if m is None:
                raise SPDXValueError("Document::Version")
            else:
                doc.version = version.Version(
                    major=int(m.group(1)), minor=int(m.group(2))
                )
                return True
        else:
            raise CardinalityError("Document::Version")

    def set_doc_data_lic(self, doc, res):
        """
        Set the document data license.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        if not self.doc_data_lics_set:
            self.doc_data_lics_set = True
            # TODO: what is this split?
            res_parts = res.split("/")
            if len(res_parts) != 0:
                identifier = res_parts[-1]
                doc.data_license = license.License.from_identifier(identifier)
            else:
                raise SPDXValueError("Document::License")
        else:
            raise CardinalityError("Document::License")

    def set_doc_name(self, doc, name):
        """
        Set the document name.
        Raise CardinalityError if already defined.
        """
        if not self.doc_name_set:
            doc.name = name
            self.doc_name_set = True
            return True
        else:
            raise CardinalityError("Document::Name")

    def set_doc_spdx_id(self, doc, doc_spdx_id_line):
        """
        Set the document SPDX Identifier.
        Raise value error if malformed value.
        Raise CardinalityError if already defined.
        """
        if not self.doc_spdx_id_set:
            if validations.validate_doc_spdx_id(doc_spdx_id_line):
                doc.spdx_id = doc_spdx_id_line
                self.doc_spdx_id_set = True
                return True
            else:
                raise SPDXValueError("Document::SPDXID")
        else:
            raise CardinalityError("Document::SPDXID")

    def set_doc_comment(self, doc, comment):
        """
        Set document comment.
        Raise CardinalityError if comment already set.
        """
        if not self.doc_comment_set:
            self.doc_comment_set = True
            doc.comment = comment
        else:
            raise CardinalityError("Document::Comment")

    def set_doc_namespace(self, doc, namespace):
        """
        Set the document namespace.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        if not self.doc_namespace_set:
            self.doc_namespace_set = True
            if validations.validate_doc_namespace(namespace):
                doc.namespace = namespace
                return True
            else:
                raise SPDXValueError("Document::Namespace")
        else:
            raise CardinalityError("Document::Comment")

    def reset_document(self):
        """
        Reset the internal state to allow building new document
        """
        # FIXME: this state does not make sense
        self.doc_version_set = False
        self.doc_comment_set = False
        self.doc_namespace_set = False
        self.doc_data_lics_set = False
        self.doc_name_set = False
        self.doc_spdx_id_set = False


class ExternalDocumentRefBuilder(tagvaluebuilders.ExternalDocumentRefBuilder):
    def set_chksum(self, doc, chk_sum):
        """
        Set the external document reference's check sum, if not already set.
        chk_sum - The checksum value in the form of a string.
        """
        if chk_sum:
            doc.ext_document_references[-1].checksum = Checksum(
                ChecksumAlgorithm.SHA1, chk_sum
            )
        else:
            raise SPDXValueError("ExternalDocumentRef::Checksum")


class EntityBuilder(tagvaluebuilders.EntityBuilder):
    def __init__(self):
        super(EntityBuilder, self).__init__()

    def create_entity(self, doc, value):
        if self.tool_re.match(value):
            return self.build_tool(doc, value)
        elif self.person_re.match(value):
            return self.build_person(doc, value)
        elif self.org_re.match(value):
            return self.build_org(doc, value)
        else:
            raise SPDXValueError("Entity")


class CreationInfoBuilder(tagvaluebuilders.CreationInfoBuilder):
    def __init__(self):
        super(CreationInfoBuilder, self).__init__()

    def set_creation_comment(self, doc, comment):
        """
        Set creation comment.
        Raise CardinalityError if comment already set.
        """
        if not self.creation_comment_set:
            self.creation_comment_set = True
            doc.creation_info.comment = comment
            return True
        else:
            raise CardinalityError("CreationInfo::Comment")


class PackageBuilder(tagvaluebuilders.PackageBuilder):
    def __init__(self):
        super(PackageBuilder, self).__init__()

    def set_pkg_checksum(self, doc, checksum: Union[Checksum, Dict]):
        """
        Set the package checksum.
        checksum - A Checksum or a Dict
        Raise SPDXValueError if checksum type invalid.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if isinstance(checksum, dict):
            algo = checksum.get('algorithm') or ChecksumAlgorithm.SHA1
            if algo.startswith('checksumAlgorithm_'):
                algo = convert_rdf_checksum_algorithm(algo) or ChecksumAlgorithm.SHA1
            else:
                algo = ChecksumAlgorithm.checksum_algorithm_from_string(algo)
            doc.packages[-1].set_checksum(Checksum(identifier=algo, value=checksum.get('checksumValue')))
        elif isinstance(checksum, Checksum):
            doc.packages[-1].set_checksum(checksum)
        elif isinstance(checksum, str):
            # kept for backwards compatibility
            doc.packages[-1].set_checksum(Checksum(identifier=ChecksumAlgorithm.SHA1, value=checksum))
        else:
            raise SPDXValueError("Invalid value for package checksum.")

    def set_pkg_source_info(self, doc, text):
        """
        Set the package's source information, if not already set.
        text - Free form text.
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not self.package_source_info_set:
            self.package_source_info_set = True
            doc.packages[-1].source_info = text
            return True
        else:
            raise CardinalityError("Package::SourceInfo")

    def set_pkg_verif_code(self, doc, code):
        """
        Set the package verification code, if not already set.
        code - A string.
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not self.package_verif_set:
            self.package_verif_set = True
            doc.packages[-1].verif_code = code
        else:
            raise CardinalityError("Package::VerificationCode")

    def set_pkg_excl_file(self, doc, filename):
        """
        Set the package's verification code excluded file.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        doc.packages[-1].add_exc_file(filename)

    def set_pkg_license_comment(self, doc, text):
        """
        Set the package's license comment.
        Raise OrderError if no package previously defined.
        Raise CardinalityError if already set.
        """
        self.assert_package_exists()
        if not self.package_license_comment_set:
            self.package_license_comment_set = True
            doc.packages[-1].license_comment = text
            return True
        else:
            raise CardinalityError("Package::LicenseComment")

    def set_pkg_attribution_text(self, doc, text):
        """
        Set the package's attribution text.
        """
        self.assert_package_exists()
        doc.packages[-1].attribution_text = text
        return True

    def set_pkg_cr_text(self, doc, text):
        """
        Set the package's license comment.
        Raise OrderError if no package previously defined.
        Raise CardinalityError if already set.
        """
        self.assert_package_exists()
        if not self.package_cr_text_set:
            self.package_cr_text_set = True
            doc.packages[-1].cr_text = text
        else:
            raise CardinalityError("Package::CopyrightText")

    def set_pkg_summary(self, doc, text):
        """
        Set the package summary.
        Raise CardinalityError if summary already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not self.package_summary_set:
            self.package_summary_set = True
            doc.packages[-1].summary = text
        else:
            raise CardinalityError("Package::Summary")

    def set_pkg_desc(self, doc, text):
        """
        Set the package's description.
        Raise CardinalityError if description already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not self.package_desc_set:
            self.package_desc_set = True
            doc.packages[-1].description = text
        else:
            raise CardinalityError("Package::Description")

    def set_pkg_comment(self, doc, text):
        """
        Set the package's comment.
        Raise CardinalityError if comment already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not self.package_comment_set:
            self.package_comment_set = True
            doc.packages[-1].comment = text
        else:
            raise CardinalityError("Package::Comment")

    def set_pkg_ext_ref_category(self, doc, category):
        """
        Set the package's external reference locator.
        Raise OrderError if no package previously defined.
        Raise SPDXValueError if malformed value.
        """
        self.assert_package_exists()
        category = category.split("_")[-1]

        if category.lower() == "packagemanager":
            category = "PACKAGE-MANAGER"

        if validations.validate_pkg_ext_ref_category(category):
            if (
                len(doc.packages[-1].pkg_ext_refs)
                and doc.packages[-1].pkg_ext_refs[-1].category is None
            ):
                doc.packages[-1].pkg_ext_refs[-1].category = category
            else:
                doc.packages[-1].add_pkg_ext_refs(
                    package.ExternalPackageRef(category=category)
                )
        else:
            raise SPDXValueError("ExternalRef::Category")

    def set_pkg_ext_ref_type(self, doc, typ):
        """
        Set the package's external reference type.
        Raise OrderError if no package previously defined.
        Raise SPDXValueError if malformed value.
        """
        self.assert_package_exists()
        if "#" in typ:
            typ = typ.split("#")[-1]
        else:
            typ = typ.split("/")[-1]

        if validations.validate_pkg_ext_ref_type(typ):
            if (
                len(doc.packages[-1].pkg_ext_refs)
                and doc.packages[-1].pkg_ext_refs[-1].pkg_ext_ref_type is None
            ):
                doc.packages[-1].pkg_ext_refs[-1].pkg_ext_ref_type = typ
            else:
                doc.packages[-1].add_pkg_ext_refs(
                    package.ExternalPackageRef(pkg_ext_ref_type=typ)
                )
        else:
            raise SPDXValueError("ExternalRef::Type")

    def set_pkg_ext_ref_comment(self, doc, comment):
        """
        Set the package's external reference comment.
        Raise CardinalityError if comment already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not len(doc.packages[-1].pkg_ext_refs):
            raise OrderError("Package::ExternalRef")
        if not self.pkg_ext_comment_set:
            self.pkg_ext_comment_set = True
            doc.packages[-1].pkg_ext_refs[-1].comment = comment
            return True
        else:
            raise CardinalityError("ExternalRef::Comment")


class FileBuilder(tagvaluebuilders.FileBuilder):
    def __init__(self):
        super(FileBuilder, self).__init__()

    def set_file_checksum(self, doc: Document, chk_sum: Union[Checksum, Dict, str]):
        """
        Set the file check sum, if not already set.
        chk_sum - A checksum.Checksum or a dict
        """
        if self.has_file(doc):
            if isinstance(chk_sum, dict):
                identifier = ChecksumAlgorithm.checksum_algorithm_from_string(chk_sum.get('algorithm'))
                self.file(doc).set_checksum(Checksum(identifier,
                                                     chk_sum.get('checksumValue')))
            elif isinstance(chk_sum, Checksum):
                self.file(doc).set_checksum(chk_sum)
            elif isinstance(chk_sum, str):
                # kept for backwards compatibility
                self.file(doc).set_checksum(Checksum(ChecksumAlgorithm.SHA1, chk_sum))
            return True

    def set_file_license_comment(self, doc, text):
        """
        Raise OrderError if no package or file defined.
        Raise CardinalityError if more than one per file.
        """
        if self.has_package(doc) and self.has_file(doc):
            if not self.file_license_comment_set:
                self.file_license_comment_set = True
                self.file(doc).license_comment = text
                return True
            else:
                raise CardinalityError("File::LicenseComment")
        else:
            raise OrderError("File::LicenseComment")

    def set_file_attribution_text(self, doc, text):
        """
        Set the file's attribution text.
        """
        if self.has_package(doc) and self.has_file(doc):
            self.assert_package_exists()
            self.file(doc).attribution_text = text
            return True

    def set_file_copyright(self, doc, text):
        """
        Raise OrderError if no package or file defined.
        Raise CardinalityError if more than one.
        """
        if self.has_package(doc) and self.has_file(doc):
            if not self.file_copytext_set:
                self.file_copytext_set = True
                self.file(doc).copyright = text
                return True
            else:
                raise CardinalityError("File::CopyRight")
        else:
            raise OrderError("File::CopyRight")

    def set_file_comment(self, doc, text):
        """
        Raise OrderError if no package or no file defined.
        Raise CardinalityError if more than one comment set.
        """
        if self.has_package(doc) and self.has_file(doc):
            if not self.file_comment_set:
                self.file_comment_set = True
                self.file(doc).comment = text
                return True
            else:
                raise CardinalityError("File::Comment")
        else:
            raise OrderError("File::Comment")

    def set_file_notice(self, doc, text):
        """
        Raise OrderError if no package or file defined.
        Raise CardinalityError if more than one.
        """
        if self.has_package(doc) and self.has_file(doc):
            if not self.file_notice_set:
                self.file_notice_set = True
                self.file(doc).notice = tagvaluebuilders.str_from_text(text)
                return True
            else:
                raise CardinalityError("File::Notice")
        else:
            raise OrderError("File::Notice")

    def set_file_type(self, doc, filetype):
        """
        Set the file type for RDF values.
        """
        if not self.has_file(doc):
            raise OrderError("File::FileType")

        split_string = filetype.split('#')
        if len(split_string) != 2:
            raise SPDXValueError('Unknown file type {}'.format(filetype))
        file_type = file.file_type_from_rdf(filetype)

        spdx_file = self.file(doc)
        if file_type in spdx_file.file_types:
            raise CardinalityError("File::FileType")

        spdx_file.file_types.append(file_type)


class SnippetBuilder(tagvaluebuilders.SnippetBuilder):
    def __init__(self):
        super(SnippetBuilder, self).__init__()

    def set_snippet_lic_comment(self, doc, lic_comment):
        """
        Set the snippet's license comment.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if already set.
        """
        self.assert_snippet_exists()
        if not self.snippet_lic_comment_set:
            self.snippet_lic_comment_set = True
            doc.snippet[-1].license_comment = lic_comment
        else:
            CardinalityError("Snippet::licenseComments")

    def set_snippet_comment(self, doc, comment):
        """
        Set general comments about the snippet.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if comment already set.
        """
        self.assert_snippet_exists()
        if not self.snippet_comment_set:
            self.snippet_comment_set = True
            doc.snippet[-1].comment = comment
            return True
        else:
            raise CardinalityError("Snippet::comment")

    def set_snippet_attribution_text(self, doc, text):
        """
        Set the snippet's attribution text.
        """
        self.assert_snippet_exists()
        doc.snippet[-1].attribution_text = text
        return True

    def set_snippet_copyright(self, doc, copyright):
        """
        Set the snippet's copyright text.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if already set.
        """
        self.assert_snippet_exists()
        if not self.snippet_copyright_set:
            self.snippet_copyright_set = True
            doc.snippet[-1].copyright = copyright
        else:
            raise CardinalityError("Snippet::copyrightText")


class ReviewBuilder(tagvaluebuilders.ReviewBuilder):
    def __init__(self):
        super(ReviewBuilder, self).__init__()

    def add_review_comment(self, doc, comment):
        """
        Set the review comment.
        Raise CardinalityError if already set.
        Raise OrderError if no reviewer defined before.
        """
        if len(doc.reviews) != 0:
            if not self.review_comment_set:
                self.review_comment_set = True
                doc.reviews[-1].comment = comment
                return True
            else:
                raise CardinalityError("ReviewComment")
        else:
            raise OrderError("ReviewComment")


class AnnotationBuilder(tagvaluebuilders.AnnotationBuilder):
    def __init__(self):
        super(AnnotationBuilder, self).__init__()

    def add_annotation_comment(self, doc, comment):
        """
        Set the annotation comment.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        """
        if len(doc.annotations) != 0:
            if not self.annotation_comment_set:
                self.annotation_comment_set = True
                doc.annotations[-1].comment = comment
                return True
            else:
                raise CardinalityError("AnnotationComment")
        else:
            raise OrderError("AnnotationComment")

    def add_annotation_type(self, doc, annotation_type):
        """
        Set the annotation type.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        """
        if len(doc.annotations) != 0:
            if not self.annotation_type_set:
                if annotation_type.endswith("annotationType_other"):
                    self.annotation_type_set = True
                    doc.annotations[-1].annotation_type = "OTHER"
                    return True
                elif annotation_type.endswith("annotationType_review"):
                    self.annotation_type_set = True
                    doc.annotations[-1].annotation_type = "REVIEW"
                    return True
                else:
                    raise SPDXValueError("Annotation::AnnotationType")
            else:
                raise CardinalityError("Annotation::AnnotationType")
        else:
            raise OrderError("Annotation::AnnotationType")


class RelationshipBuilder(tagvaluebuilders.RelationshipBuilder):
    def __init__(self):
        super(RelationshipBuilder, self).__init__()

    def add_relationship_comment(self, doc, comment):
        """
        Set the relationship comment.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        """
        if len(doc.relationships) != 0:
            if not self.relationship_comment_set:
                self.relationship_comment_set = True
                doc.relationships[-1].comment = comment
                return True
            else:
                raise CardinalityError("RelationshipComment")
        else:
            raise OrderError("RelationshipComment")


class Builder(
    DocBuilder,
    EntityBuilder,
    CreationInfoBuilder,
    PackageBuilder,
    FileBuilder,
    SnippetBuilder,
    ReviewBuilder,
    ExternalDocumentRefBuilder,
    AnnotationBuilder,
    RelationshipBuilder,
):
    def __init__(self):
        super(Builder, self).__init__()
        # FIXME: this state does not make sense
        self.reset()

    def reset(self):
        """
        Reset builder's state for building new documents.
        Must be called between usage with different documents.
        """
        # FIXME: this state does not make sense
        self.reset_creation_info()
        self.reset_document()
        self.reset_package()
        self.reset_file_stat()
        self.reset_reviews()
        self.reset_annotations()
        self.reset_relationship()



#!/usr/bin/env python

# Copyright (c) 2020 Yash Varshney
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spdx.parsers.builderexceptions import FileTypeError
from spdx.parsers.parse_anything import parse_file
from spdx.writers.write_anything import write_file

import click

def print_help_msg(command):
    with click.Context(command) as ctx:
        click.echo(command.get_help(ctx))

def determine_infile_and_outfile(infile, outfile, src, from_, to):
    if infile is not None and outfile is not None:
        """
        when the CLI is of given format:
        ' pyspdxtools_convertor ---infile <input_file> ---outfile <output_file>.
        """
        return infile, outfile

    elif infile is None and outfile is None and len(src) == 2:
        """
        ' pyspdxtools_convertor -f/--from <type> <input_file> -t/--to <type> <output_file>.
        """
        infile = src[0]
        outfile = src[1]
        # infile = os.path.splitext(infile)[0]
        if from_ is not None:
            infile_path = os.path.splitext(infile)[0]
            infile = infile_path + "." + from_
        if to is not None:
            outfile_path = os.path.splitext(outfile)[0]
            outfile = outfile_path + "." + to
        return infile, outfile

    elif infile is None and outfile is not None:
        """
        ' pyspdxtools_convertor -f/--from <type> <input_file> --outfile <output_file> '
        """
        infile = src[0]
        if from_ is not None:
            infile_path = os.path.splitext(infile)[0]
            infile = infile_path + "." + from_
        return infile, outfile

    elif infile is not None and outfile is None:
        """
        ' pyspdxtools_convertor --infile <input_file> -t/--to <type> <output_file>'
        """
        outfile = src[0]
        if to is not None:
            outfile_path = os.path.splitext(outfile)[0]
            outfile = outfile_path + "." + to
        return infile, outfile

    else:
        raise ValueError("Given arguments for convertor are invalid.")


@click.command()
@click.argument("src", nargs=-1)
@click.option("--infile", "-i", help="The file to be converted ")
@click.option("--outfile", "-o", help="The file after converting")
@click.option(
    "--to",
    "-t",
    type=click.Choice(["json", "rdf", "yaml", "xml", "tag"], case_sensitive=False)
)
@click.option(
    "--from",
    "-f",
    "from_",
    type=click.Choice(["tag", "rdf"], case_sensitive=False))
@click.option("--force", is_flag=True, help="convert even if there are some parsing errors or inconsistencies")
def main(infile, outfile, src, from_, to, force):
    """
    CLI-TOOL for converting a RDF or TAG file to RDF, JSON, YAML, TAG or XML format.

    To use : run 'pyspdxtools_convertor -f <from_TYPE> <input file> -t <to_TYPE> <output_file>' command on terminal or use ' pyspdxtools_convertor --infile <input file name> --outfile <output file name> '

    """
    try:
        infile, outfile = determine_infile_and_outfile(infile, outfile, src, from_, to)
    except ValueError as err:
        print(err)
        print_help_msg(main)
        return

    doc, errors = parse_file(infile)
    if errors:
        print("Errors while parsing: ", errors)
        if not force:
            return 1

    write_file(doc, outfile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (C) 2017 BMW AG
# Author: Thomas Hafner
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from spdx.parsers.loggers import StandardLogger
from spdx.writers.rdf import write_document
from spdx.parsers.tagvalue import Parser
from spdx.parsers.tagvaluebuilders import Builder
from spdx.parsers.loggers import ErrorMessages


def tv_to_rdf(infile_name, outfile_name):
    """
    Convert a SPDX file from tag/value format to RDF format.
    Return True on success, False otherwise.
    """
    parser = Parser(Builder(), StandardLogger())
    parser.build()
    with open(infile_name) as infile:
        data = infile.read()
        document, error = parser.parse(data)
        if not error:
            with open(outfile_name, mode="wb") as outfile:
                write_document(document, outfile)
            return True
        else:
            print("Errors encountered while parsing RDF file.")
            messages = ErrorMessages()
            document.validate(messages)
            print("\n".join(messages.messages))
            return False


def main():
    args = sys.argv[1:]
    if not args:
        print(
            "Usage: spdx-tv2rdf <tag-value-file> <rdf-file>\n"
            "Convert an SPDX tag/value document to RDF."
        )
        sys.exit(1)

    tv_file = args[0]
    rdf_file = args[1]
    success = tv_to_rdf(tv_file, rdf_file)
    sys.exit(0 if success else 1)


if __name__ == "__main__":
    main()



# Copyright (c) the SPDX tools authors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from spdx import config
from spdx.version import Version


class TestLicenseList(TestCase):
    maxDiff = None

    def test_load_license_list(self):
        version, licenses_map = config.load_license_list(config._licenses)
        assert version == ('3', '20')
        # Test some instances in licenses_map
        assert licenses_map['MIT License'] == 'MIT'
        assert licenses_map['MIT'] == 'MIT License'
        assert licenses_map['Apache License 2.0'] == 'Apache-2.0'
        assert licenses_map['Apache-2.0'] == 'Apache License 2.0'
        assert licenses_map['GNU General Public License v3.0 only'] == 'GPL-3.0-only'
        assert licenses_map['GPL-3.0-only'] == 'GNU General Public License v3.0 only'

    def test_config_license_list_version_constant(self):
        assert config.LICENSE_LIST_VERSION == Version(major=3, minor=20)

    def test_load_exception_list(self):
        version, exception_map = config.load_exception_list(config._exceptions)
        assert version == ('3', '20')
        # Test some instances in exception_map
        assert exception_map['Bison exception 2.2'] == 'Bison-exception-2.2'
        assert exception_map['Bison-exception-2.2'] == 'Bison exception 2.2'
        assert exception_map['OpenVPN OpenSSL Exception'] == 'openvpn-openssl-exception'
        assert exception_map['openvpn-openssl-exception'] == 'OpenVPN OpenSSL Exception'
        assert exception_map['Qt GPL exception 1.0'] == 'Qt-GPL-exception-1.0'
        assert exception_map['Qt-GPL-exception-1.0'] == 'Qt GPL exception 1.0'



# Copyright (c) 2021 spdx tool contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from spdx.parsers import parse_anything
from spdx.writers import write_anything
from tests import utils_test

dirname = os.path.join(os.path.dirname(__file__), "data", "formats")
test_files = [os.path.join(dirname, fn) for fn in os.listdir(dirname)]
test_files_json_yaml_xml_tag = [filename for filename in test_files if filename.endswith("json")
                                or filename.endswith("yaml") or filename.endswith("xml") or filename.endswith("tag")]
test_files_rdf = [filename for filename in test_files if filename.endswith("rdf")]
UNSTABLE_CONVERSIONS = {
    "SPDXTagExample.tag-yaml",
    "SPDXTagExample.tag-xml",
    "SPDXTagExample.tag-json",
    "SPDXXmlExample.xml-tag",
    "SPDXJsonExample.json-tag",
    "SPDXYamlExample.yaml-tag",
    "SPDXRdfExample.rdf-rdf",
    "SPDXYAMLExample-2.2.spdx.yaml-tag",
    "SPDXJSONExample-v2.2.spdx.json-tag",
    "SPDXXMLExample-v2.2.spdx.xml-tag",
    "SPDXYAMLExample-2.3.spdx.yaml-tag",
    "SPDXJSONExample-v2.3.spdx.json-tag",
    "SPDXXMLExample-v2.3.spdx.xml-tag"
}


# Because the rdf-parser/ writer can't handle the mandatory field byte_range in snippets yet we can only test conversion
# from json, yaml, xml and tv to each other format and rdf to rdf. Otherwise, the jsonyamlxml- or tv-writer would add
# the initial value None for snippet_ranges which then leads to an error while parsing.
# https://github.com/spdx/tools-python/issues/274


@pytest.mark.parametrize("out_format", ['yaml', 'xml', 'json', 'tag'])
@pytest.mark.parametrize("in_file", test_files_json_yaml_xml_tag, ids=lambda x: os.path.basename(x))
def test_write_anything_json_yaml_xml_tv(in_file, out_format, tmpdir):
    in_basename = os.path.basename(in_file)
    write_anything_test(in_basename, in_file, out_format, tmpdir)


@pytest.mark.parametrize("out_format", ['rdf'])
@pytest.mark.parametrize("in_file", test_files_rdf, ids=lambda x: os.path.basename(x))
def test_write_anything_rdf(in_file, out_format, tmpdir):
    in_basename = os.path.basename(in_file)
    write_anything_test(in_basename, in_file, out_format, tmpdir)


def write_anything_test(in_basename, in_file, out_format, tmpdir):
    """This parses the in_file and writes it to the out_format,
    then parses the written out_file again and checks if it is still the same as in_file."""
    doc_in, error_in = parse_anything.parse_file(in_file)
    assert not error_in
    result_in = utils_test.TestParserUtils.to_dict(doc_in)

    out_file_name = os.path.join(tmpdir, "test." + out_format)
    write_anything.write_file(doc_in, out_file_name)

    doc_out, error_out = parse_anything.parse_file(out_file_name)
    assert not error_out
    result_out = utils_test.TestParserUtils.to_dict(doc_out)

    test = in_basename + "-" + out_format
    if test not in UNSTABLE_CONVERSIONS:
        assert result_in == result_out

    else:
        # if this test fails, this means we are more stable \o/
        # in that case, please remove the test from UNSTABLE_CONVERSIONS list
        assert result_out != result_in, test



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from datetime import datetime
from enum import Enum
from functools import reduce
from typing import Optional, Union

from spdx import creationinfo
from spdx import license
from spdx import utils
from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.creationinfo import Creator
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.parsers.loggers import ErrorMessages
from spdx.utils import NoAssert


class PackagePurpose(Enum):
    APPLICATION = 1
    FRAMEWORK = 2
    LIBRARY = 3
    CONTAINER = 4
    OPERATING_SYSTEM = 5
    DEVICE = 6
    FIRMWARE = 7
    SOURCE = 8
    ARCHIVE = 9
    FILE = 10
    INSTALL = 11
    OTHER = 12


class Package(object):
    """
    Represent an analyzed Package.
    Fields:
     - name: Mandatory, string.
     - spdx_id: Uniquely identify any element in an SPDX document which may be
     referenced by other elements. Mandatory, one. Type: str.
     - version: Optional, string.
     - file_name: Optional, string.
     - supplier: Optional, Organization or Person or NO_ASSERTION.
     - originator: Optional, Organization or Person.
     - download_location: Mandatory, URL as string.
     - files_analyzed: Indicates whether the file content of this package has
     been available for or subjected to analysis when creating the SPDX
     document. If "false" indicates packages that represent metadata or URI
     references to a project, product, artifact, distribution or a component.
     If set to "false", the package must not contain any files.
     Optional, boolean.
     - homepage: Optional, URL as string or NONE or NO_ASSERTION.
     - verif_code: string. According to the specification, this is Mandatory
     whenever files_analyzed is True or None (omitted) and Must be None (omitted)
     if files_analyzed is False. However, as a convenience within this library,
     we allow this to be Optional even when files_analyzed is True/None.
     - checksums: Optional, Dict with checksum.ChecksumAlgorithm as key and checksum.Checksum as value.
     - source_info: Optional string.
     - conc_lics: Mandatory license.License or utils.SPDXNone or
     utils.NoAssert.
     - license_declared: Mandatory license.License or utils.SPDXNone or
     utils.NoAssert.
     - license_comment: optional string.
     - licenses_from_files: list of license.License or utils.SPDXNone or
     utils.NoAssert.
     - cr_text: Copyright text, string , utils.NoAssert or utils.SPDXNone. Mandatory.
     - summary: Optional str.
     - description: Optional str.
     - comment: Comments about the package being described, optional one.
     Type: str
     - verif_exc_files: list of file names excluded from verification code or None.
     - ext_pkg_refs: External references referenced within the given package.
     Optional, one or many. Type: ExternalPackageRef
     - attribution_text: optional string.
     - primary_package_purpose: Optional one. Type: PackagePurpose
    """

    def __init__(
        self,
        name=None,
        spdx_id=None,
        download_location=None,
        version=None,
        file_name=None,
        supplier=None,
        originator=None,
    ):
        self.name = name
        self.spdx_id = spdx_id
        self.version = version
        self.file_name = file_name
        self.supplier: Optional[Union[Creator, NoAssert]] = supplier
        self.originator: Optional[Union[Creator, NoAssert]] = originator
        self.download_location = download_location
        self.files_analyzed = None
        self.homepage = None
        self.verif_code = None
        self.checksums = {}
        self.source_info = None
        self.conc_lics = None
        self.license_declared = None
        self.license_comment = None
        self.licenses_from_files = []
        self.cr_text = None
        self.summary = None
        self.description = None
        self.comment = None
        self.attribution_text = None
        self.verif_exc_files = []
        self.pkg_ext_refs = []
        self.primary_package_purpose: Optional[PackagePurpose] = None
        self.release_date: Optional[datetime] = None
        self.built_date: Optional[datetime] = None
        self.valid_until_date: Optional[datetime] = None

    @property
    def checksum(self):
        """
        Backwards compatibility, return SHA1 checksum.
        """
        warnings.warn("This property is deprecated. Use get_checksum instead.")
        return self.get_checksum(ChecksumAlgorithm.SHA1)

    @checksum.setter
    def checksum(self, value):
        """
        Backwards compatibility, set SHA1 checksum.
        """
        warnings.warn("This property is deprecated. Use set_checksum instead.")
        if isinstance(value, str):
            self.set_checksum(Checksum("SHA1", value))
        elif isinstance(value, Checksum):
            self.set_checksum(value)

    @property
    def are_files_analyzed(self):
        return self.files_analyzed is not False
        # as default None Value is False, previous line is simplification of
        # return self.files_analyzed or self.files_analyzed is None

    def add_lics_from_file(self, lics):
        self.licenses_from_files.append(lics)

    def add_exc_file(self, filename):
        self.verif_exc_files.append(filename)

    def add_pkg_ext_refs(self, pkg_ext_ref):
        self.pkg_ext_refs.append(pkg_ext_ref)

    def validate(self, messages):
        """
        Validate the package fields.
        Append user friendly error messages to the `messages` list.
        """
        messages.push_context(self.name)
        self.validate_files_analyzed(messages)
        self.validate_checksums(messages)
        self.validate_optional_str_fields(messages)
        self.validate_mandatory_str_fields(messages)
        self.validate_pkg_ext_refs(messages)
        self.validate_optional_fields(messages)
        messages.pop_context()

        return messages

    def validate_files_analyzed(self, messages):
        if self.files_analyzed not in [True, False, None]:
            messages.append(
                'Package files_analyzed must be True or False or None (omitted)'
            )
        if not self.are_files_analyzed and self.verif_code is not None:
            messages.append(
                'Package verif_code must be None (omitted) when files_analyzed is False'
            )

        return messages

    def validate_primary_package_purposes(self, messages: ErrorMessages) -> ErrorMessages:
        if self.primary_package_purpose not in PackagePurpose:
            messages.append("Primary package purpose has a value that is not allowed!")
        return messages

    def validate_optional_fields(self, messages):
        if self.originator and not isinstance(
            self.originator, (utils.NoAssert, creationinfo.Creator)
        ):
            messages.append(
                "Package originator must be instance of "
                "spdx.utils.NoAssert or spdx.creationinfo.Creator"
            )

        if self.supplier and not isinstance(
            self.supplier, (utils.NoAssert, creationinfo.Creator)
        ):
            messages.append(
                "Package supplier must be instance of "
                "spdx.utils.NoAssert or spdx.creationinfo.Creator"
            )

        if self.conc_lics and not isinstance(
            self.conc_lics, (utils.SPDXNone, utils.NoAssert, license.License)
        ):
            messages.append(
                "Package concluded license must be instance of "
                "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                "spdx.license.License"
            )

        if self.license_declared and not isinstance(
            self.license_declared, (utils.SPDXNone, utils.NoAssert, license.License)
        ):
            messages.append(
                "Package declared license must be instance of "
                "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                "spdx.license.License"
            )

        license_from_file_check = lambda prev, el: prev and isinstance(
            el, (license.License, utils.SPDXNone, utils.NoAssert)
        )
        if not reduce(license_from_file_check, self.licenses_from_files, True):
            messages.append(
                "Each element in licenses_from_files must be instance of "
                "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                "spdx.license.License"
            )

        return messages

    def validate_pkg_ext_refs(self, messages):
        for ref in self.pkg_ext_refs:
            if isinstance(ref, ExternalPackageRef):
                messages = ref.validate(messages)
            else:
                messages.append(
                    "External package references must be of the type "
                    "spdx.package.ExternalPackageRef and not " + str(type(ref))
                )

        return messages

    def validate_optional_str_fields(self, messages):
        """Fields marked as optional and of type string in class
        docstring must be of a type that provides __str__ method.
        """
        FIELDS = [
            "file_name",
            "version",
            "homepage",
            "source_info",
            "summary",
            "description",
            "attribution_text",
            "comment",
            "cr_text"
        ]
        self.validate_str_fields(FIELDS, True, messages)

        return messages

    def validate_mandatory_str_fields(self, messages):
        """Fields marked as Mandatory and of type string in class
        docstring must be of a type that provides __str__ method.
        """
        FIELDS = ["name", "spdx_id", "download_location"]
        self.validate_str_fields(FIELDS, False, messages)

        return messages

    def validate_str_fields(self, fields, optional, messages):
        """Helper for validate_mandatory_str_field and
        validate_optional_str_fields"""
        for field_str in fields:
            field = getattr(self, field_str)
            if field is not None:
                # FIXME: this does not make sense???
                attr = getattr(field, "__str__", None)
                if not callable(attr):
                    messages.append(
                        "{0} must provide __str__ method.".format(field)
                    )
                    # Continue checking.
            elif not optional:
                messages.append("Package {0} can not be None.".format(field_str))

        return messages

    def validate_checksums(self, messages: ErrorMessages):
        if not self.checksums:
            return
        for checksum in self.checksums.values():
            if not isinstance(checksum, Checksum):
                messages.append("Package checksum must be instance of spdx.checksum.Checksum")

    def get_checksum(self, hash_algorithm: ChecksumAlgorithm = ChecksumAlgorithm.SHA1) -> Optional[Checksum]:
        return self.checksums.get(hash_algorithm)

    def set_checksum(self, new_checksum: Checksum):
        if not isinstance(new_checksum, Checksum):
            raise SPDXValueError("Package::Checksum")

        self.checksums[new_checksum.identifier] = new_checksum

    def has_optional_field(self, field):
        return bool(getattr(self, field, None))


class ExternalPackageRef(object):
    """
    An External Reference allows a Package to reference an external source of
    additional information, metadata, enumerations, asset identifiers, or
    downloadable content believed to be relevant to the Package.
    Fields:
    - category: "SECURITY" or "PACKAGE-MANAGER" or "OTHER".
    - pkg_ext_ref_type: A unique string containing letters, numbers, ".","-".
    - locator: A unique string with no spaces necessary to access the
    package-specific information, metadata, or content within the target
    location.
    - comment: To provide information about the purpose and target of the
    reference.
    """

    def __init__(
        self, category=None, pkg_ext_ref_type=None, locator=None, comment=None
    ):
        self.category = category
        self.pkg_ext_ref_type = pkg_ext_ref_type
        self.locator = locator
        self.comment = comment

    def validate(self, messages):
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        self.validate_category(messages)
        self.validate_pkg_ext_ref_type(messages)
        self.validate_locator(messages)

        return messages

    def validate_category(self, messages):
        if self.category is None:
            messages.append("ExternalPackageRef has no category.")

        return messages

    def validate_pkg_ext_ref_type(self, messages):
        if self.pkg_ext_ref_type is None:
            messages.append("ExternalPackageRef has no type.")

        return messages

    def validate_locator(self, messages):
        if self.locator is None:
            messages.append("ExternalPackageRef has no locator.")

        return messages



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import TestCase

from spdx.checksum import Checksum
from spdx.package import Package


class TestPackage(TestCase):

    @unittest.skip("https://github.com/spdx/tools-python/issues/296")
    def test_calc_verif_code(self):
        package = Package()
        package.calc_verif_code()

    def test_package_with_non_sha1_check_sum(self):
        package = Package()
        package.set_checksum(Checksum("SHA256", ''))

        # Make sure that validation still works despite the checksum not being SHA1
        messages = []
        package.validate_checksums(messages)


if __name__ == '__main__':
    unittest.main()
